from __future__ import annotations

import functools
import operator
import os
import re
from pathlib import Path
from typing import TYPE_CHECKING, Any, Iterable, Mapping, cast

from pdm.compat import tomllib
from pdm.formats.base import (
    MetaConverter,
    Unset,
    array_of_inline_tables,
    convert_from,
    make_array,
    make_inline_table,
)
from pdm.models.markers import Marker, get_marker
from pdm.models.requirements import FileRequirement, Requirement
from pdm.models.specifiers import PySpecSet
from pdm.utils import cd

if TYPE_CHECKING:
    from argparse import Namespace

    from pdm._types import RequirementDict
    from pdm.project import Project


def check_fingerprint(project: Project | None, filename: Path | str) -> bool:
    if Path(filename).name != "pyproject.toml":
        return False
    with open(filename, "rb") as fp:
        try:
            data = tomllib.load(fp)
        except tomllib.TOMLDecodeError:
            return False

    return "tool" in data and "poetry" in data["tool"]


VERSION_RE = re.compile(r"([^\d\s]*)\s*(\d.*?)\s*(?=,|$)")


def _convert_specifier(version: str) -> str:
    parts = []
    for op, ver in VERSION_RE.findall(str(version)):
        if op == "~":
            op += "="
        elif op == "^":
            major, *vparts = ver.split(".")
            next_major = ".".join([str(int(major) + 1)] + ["0"] * len(vparts))
            parts.append(f">={ver},<{next_major}")
            continue
        elif not op:
            op = "=="
        parts.append(f"{op}{ver}")
    return ",".join(parts)


def _convert_python(python: str) -> PySpecSet:
    if not python:
        return PySpecSet()
    parts = [PySpecSet(_convert_specifier(s)) for s in python.split("||")]
    return functools.reduce(operator.or_, parts)


def _convert_req(name: str, req_dict: RequirementDict | list[RequirementDict]) -> Iterable[str]:
    from pdm.models.backends import DEFAULT_BACKEND

    backend = DEFAULT_BACKEND(Path.cwd())

    def fix_req_path(req: Requirement) -> Requirement:
        if isinstance(req, FileRequirement):
            req.relocate(backend)
        return req

    if isinstance(req_dict, list):
        for req in req_dict:
            yield from _convert_req(name, req)
    elif isinstance(req_dict, str):
        pdm_req = fix_req_path(Requirement.from_req_dict(name, _convert_specifier(req_dict)))
        yield pdm_req.as_line()
    else:
        assert isinstance(req_dict, dict)
        req_dict = dict(req_dict)
        req_dict.pop("optional", None)  # Ignore the 'optional' key
        if "version" in req_dict:
            req_dict["version"] = _convert_specifier(str(req_dict["version"]))
        markers: list[Marker] = []
        if "markers" in req_dict:
            markers.append(get_marker(req_dict.pop("markers")))  # type: ignore[arg-type]
        if "python" in req_dict:
            markers.append(get_marker(_convert_python(str(req_dict.pop("python"))).as_marker_string()))
        if markers:
            req_dict["marker"] = str(functools.reduce(operator.and_, markers)).replace('"', "'")
        if "rev" in req_dict or "branch" in req_dict or "tag" in req_dict:
            req_dict["ref"] = req_dict.pop(
                "rev",
                req_dict.pop("tag", req_dict.pop("branch", None)),  # type: ignore[arg-type]
            )
        pdm_req = fix_req_path(Requirement.from_req_dict(name, req_dict))
        yield pdm_req.as_line()


# See https://github.com/python-poetry/poetry-core/pull/521#issuecomment-1327689551
# for reasoning why email.utils.parseaddr is not used here.
NAME_EMAIL_RE = re.compile(
    r"^(?P<name>[- .,\w'’\"():&]+)(?: <(?P<email>.+?)>)?$",  # noqa: RUF001
    re.UNICODE,
)


def parse_name_email(name_email: list[str]) -> list[dict]:
    return array_of_inline_tables(
        [
            {
                k: v
                for k, v in NAME_EMAIL_RE.match(item).groupdict().items()  # type: ignore[union-attr]
                if v is not None
            }
            for item in name_email
        ]
    )


class PoetryMetaConverter(MetaConverter):
    @convert_from("authors")
    def authors(self, value: list[str]) -> list[dict]:
        return parse_name_email(value)

    @convert_from("maintainers")
    def maintainers(self, value: list[str]) -> list[dict]:
        return parse_name_email(value)

    @convert_from("license")
    def license(self, value: str) -> dict[str, str]:
        return make_inline_table({"text": value})

    @convert_from(name="requires-python")
    def requires_python(self, source: dict[str, Any]) -> str:
        python = source.get("dependencies", {}).pop("python", None)
        return str(_convert_python(python))

    @convert_from()
    def urls(self, source: dict[str, Any]) -> dict[str, str]:
        rv = source.pop("urls", {})
        if "homepage" in source:
            rv["homepage"] = source.pop("homepage")
        if "repository" in source:
            rv["repository"] = source.pop("repository")
        if "documentation" in source:
            rv["documentation"] = source.pop("documentation")
        return rv

    @convert_from("plugins", name="entry-points")
    def entry_points(self, value: dict[str, dict[str, str]]) -> dict[str, dict[str, str]]:
        return value

    @convert_from()
    def dependencies(self, source: dict[str, Any]) -> list[str]:
        rv = []
        value, extras = dict(source["dependencies"]), source.pop("extras", {})
        for key, req_dict in value.items():
            optional = getattr(req_dict, "items", None) and req_dict.pop("optional", False)
            for req in _convert_req(key, req_dict):
                if optional:
                    extra = next((k for k, v in extras.items() if key in v), None)
                    if extra:
                        self._data.setdefault("optional-dependencies", {}).setdefault(extra, []).append(req)
                else:
                    rv.append(req)
        del source["dependencies"]
        return make_array(rv, True)

    @convert_from("dev-dependencies")
    def dev_dependencies(self, value: dict) -> None:
        self.settings.setdefault("dev-dependencies", {})["dev"] = make_array(
            [r for key, req in value.items() for r in _convert_req(key, req)], True
        )
        raise Unset()

    @convert_from("group")
    def group_dependencies(self, value: dict[str, dict[str, Any]]) -> None:
        for name, group in value.items():
            self.settings.setdefault("dev-dependencies", {})[name] = make_array(
                [r for key, req in group.get("dependencies", {}).items() for r in _convert_req(key, req)], True
            )
        raise Unset()

    @convert_from("package-mode")
    def package_mode(self, value: bool) -> None:
        self.settings["distribution"] = value
        raise Unset()

    @convert_from()
    def includes(self, source: dict[str, list[str] | str]) -> list[str]:
        includes: list[str] = []
        source_includes: list[str] = []
        for item in source.pop("packages", []):
            assert isinstance(item, dict)
            include = item["include"]
            if item.get("from"):
                include = Path(str(item.get("from")), include).as_posix()
            includes.append(include)
        for item in source.pop("include", []):
            if not isinstance(item, dict):
                includes.append(item)
            else:
                dest = source_includes if "sdist" in item.get("format", "") else includes
                dest.append(item["path"])
        if includes:
            self.settings.setdefault("build", {})["includes"] = includes
        raise Unset()

    @convert_from("exclude")
    def excludes(self, value: list[str]) -> None:
        self.settings.setdefault("build", {})["excludes"] = value
        raise Unset()

    @convert_from("build")
    def build(self, value: str | dict) -> None:
        result = {}
        if isinstance(value, dict) and "generate-setup-file" in value:
            result["run-setuptools"] = cast(bool, value["generate-setup-file"])
        self.settings.setdefault("build", {}).update(result)
        raise Unset()

    @convert_from("source")
    def sources(self, value: list[dict[str, Any]]) -> None:
        self.settings["source"] = [
            {
                "name": item.get("name", ""),
                "url": item.get("url", ""),
                "verify_ssl": item.get("url", "").startswith("https"),
            }
            for item in value
        ]
        raise Unset()


def convert(
    project: Project | None,
    filename: str | Path,
    options: Namespace | None,
) -> tuple[Mapping[str, Any], Mapping[str, Any]]:
    with open(filename, "rb") as fp, cd(os.path.dirname(os.path.abspath(filename))):
        converter = PoetryMetaConverter(tomllib.load(fp)["tool"]["poetry"], project.core.ui if project else None)
        return converter.convert()


def export(project: Project, candidates: list, options: Any) -> None:
    raise NotImplementedError()



from pdm.builders.editable import EditableBuilder
from pdm.builders.sdist import SdistBuilder
from pdm.builders.wheel import WheelBuilder

__all__ = ["EditableBuilder", "SdistBuilder", "WheelBuilder"]



from __future__ import annotations

import dataclasses
import hashlib
import os
import re
import warnings
from functools import cached_property
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import TYPE_CHECKING, Any, ClassVar, cast, no_type_check
from zipfile import ZipFile

from packaging.version import InvalidVersion

from pdm import termui
from pdm.builders import EditableBuilder, WheelBuilder
from pdm.compat import importlib_metadata as im
from pdm.exceptions import BuildError, CandidateNotFound, InvalidPyVersion, PDMWarning, RequirementError
from pdm.models.backends import get_backend, get_backend_by_spec
from pdm.models.reporter import CandidateReporter
from pdm.models.requirements import (
    FileRequirement,
    Requirement,
    VcsRequirement,
    _egg_info_re,
    filter_requirements_with_extras,
)
from pdm.models.setup import Setup
from pdm.models.specifiers import PySpecSet
from pdm.utils import (
    comparable_version,
    convert_hashes,
    filtered_sources,
    get_rev_from_url,
    normalize_name,
    url_without_fragments,
)

if TYPE_CHECKING:
    from importlib.metadata import _SimplePath

    from unearth import Link, Package, PackageFinder

    from pdm._types import FileHash
    from pdm.environments import BaseEnvironment


def _dist_info_files(whl_zip: ZipFile) -> list[str]:
    """Identify the .dist-info folder inside a wheel ZipFile."""
    res = []
    for path in whl_zip.namelist():
        m = re.match(r"[^/\\]+-[^/\\]+\.dist-info/", path)
        if m:
            res.append(path)
    if res:
        return res
    raise Exception("No .dist-info folder found in wheel")


def _get_wheel_metadata_from_wheel(whl_file: Path, metadata_directory: str) -> str:
    """Extract the metadata from a wheel.
    Fallback for when the build backend does not
    define the 'get_wheel_metadata' hook.
    """
    with ZipFile(whl_file) as zipf:
        dist_info = _dist_info_files(zipf)
        zipf.extractall(path=metadata_directory, members=dist_info)
    return os.path.join(metadata_directory, dist_info[0].split("/")[0])


def _filter_none(data: dict[str, Any]) -> dict[str, Any]:
    """Return a new dict without None values"""
    return {k: v for k, v in data.items() if v is not None}


def _find_best_match_link(finder: PackageFinder, req: Requirement, files: list[FileHash]) -> Link | None:
    """Get the best matching link for a requirement"""

    # This function is called when a lock file candidate is given or incompatible wheel
    # In this case, the requirement must be pinned, so no need to pass allow_prereleases
    # If links are not empty, find the best match from the links, otherwise find from
    # the package sources.
    from unearth import Link

    links = [Link(f["url"]) for f in files if "url" in f]
    hashes = convert_hashes(files)

    if not links:
        best = finder.find_best_match(req.as_line(), hashes=hashes).best
    else:
        # this branch won't be executed twice if ignore_compatibility is True
        evaluator = finder.build_evaluator(req.name)
        packages = finder._evaluate_links(links, evaluator)
        best = max(packages, key=finder._sort_key, default=None)
    return best.link if best is not None else None


class MetadataDistribution(im.Distribution):
    """A wrapper around a single METADATA file to provide the Distribution interface"""

    def __init__(self, text: str) -> None:
        self.text = text

    def locate_file(self, path: str | os.PathLike[str]) -> _SimplePath:
        return Path()

    def read_text(self, filename: str) -> str | None:
        if filename != "":
            return None
        return self.text


class Candidate:
    """A concrete candidate that can be downloaded and installed.
    A candidate comes from the PyPI index of a package, or from the requirement itself
    (for file or VCS requirements). Each candidate has a name, version and several
    dependencies together with package metadata.
    """

    __slots__ = (
        "_preferred",
        "_prepared",
        "_requires_python",
        "_revision",
        "hashes",
        "installed",
        "link",
        "name",
        "req",
        "requested",
        "summary",
        "version",
    )

    def __init__(
        self,
        req: Requirement,
        name: str | None = None,
        version: str | None = None,
        link: Link | None = None,
        installed: im.Distribution | None = None,
    ):
        """
        :param req: the requirement that produces this candidate.
        :param name: the name of the candidate.
        :param version: the version of the candidate.
        :param link: the file link of the candidate.
        """
        self.req = req
        self.name = name or self.req.project_name
        self.version = version
        if link is None and not req.is_named:
            link = cast("Link", req.as_file_link())  # type: ignore[attr-defined]
        self.link = link
        self.summary = ""
        self.hashes: list[FileHash] = []
        self.requested = False
        self.installed: im.Distribution | None = installed

        self._requires_python: str | None = None
        self._prepared: PreparedCandidate | None = None
        self._revision = getattr(req, "revision", None)

    def identify(self) -> str:
        return self.req.identify()

    def copy_with(self, requirement: Requirement) -> Candidate:
        can = Candidate(requirement, name=self.name, version=self.version, link=self.link, installed=self.installed)
        can.summary = self.summary
        can.hashes = self.hashes
        can._requires_python = self._requires_python
        can._prepared = self._prepared
        can._revision = self._revision
        if can._prepared:
            can._prepared.req = requirement
        return can

    @property
    def dep_key(self) -> tuple[str, str | None]:
        """Key for retrieving and storing dependencies from the provider.

        Return a tuple of (name, version). For URL candidates, the version is None but
        there will be only one for the same name so it is also unique.
        """
        return (self.identify(), self.version)

    @property
    def prepared(self) -> PreparedCandidate | None:
        return self._prepared

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Candidate):
            return False
        if self.req.is_named:
            return self.name == other.name and self.version == other.version
        return self.name == other.name and self.link == other.link

    def get_revision(self) -> str:
        if not self.req.is_vcs:
            raise AttributeError("Non-VCS candidate doesn't have revision attribute")
        if self._revision:
            return self._revision
        if self.req.revision:  # type: ignore[attr-defined]
            return self.req.revision  # type: ignore[attr-defined]
        return self._prepared.revision if self._prepared else "unknown"

    def __repr__(self) -> str:
        source = getattr(self.link, "comes_from", None)
        from_source = f" from {source}" if source else ""
        return f"<Candidate {self}{from_source}>"

    def __str__(self) -> str:
        if self.req.is_named:
            return f"{self.name}@{self.version}"
        assert self.link is not None
        return f"{self.name}@{self.link.url_without_fragment}"

    @classmethod
    def from_installation_candidate(cls, candidate: Package, req: Requirement) -> Candidate:
        """Build a candidate from unearth's find result."""
        return cls(
            req,
            name=candidate.name,
            version=str(candidate.version),
            link=candidate.link,
        )

    @property
    def requires_python(self) -> str:
        """The Python version constraint of the candidate."""
        if self._requires_python is not None:
            return self._requires_python
        if self.link:
            requires_python = self.link.requires_python
            if requires_python is not None:
                if requires_python.isdigit():
                    requires_python = f">={requires_python},<{int(requires_python) + 1}"
                try:  # ensure the specifier is valid
                    PySpecSet(requires_python)
                except InvalidPyVersion:
                    pass
                else:
                    self._requires_python = requires_python
        return self._requires_python or ""

    @requires_python.setter
    def requires_python(self, value: str) -> None:
        try:  # ensure the specifier is valid
            PySpecSet(value)
        except InvalidPyVersion:
            return
        self._requires_python = value

    @no_type_check
    def as_lockfile_entry(self, project_root: Path) -> dict[str, Any]:
        """Build a lockfile entry dictionary for the candidate."""
        version = str(self.version)
        if self.req.is_pinned:
            spec = next(iter(self.req.specifier))
            should_normalize_version = "+" not in spec.version
        else:
            should_normalize_version = True

        if should_normalize_version:
            try:
                version = str(comparable_version(version))
            except InvalidVersion as e:
                raise RequirementError(f"Invalid version for {self.req.as_line()}: {e}") from None
        result = {
            "name": normalize_name(self.name),
            "version": version,
            "extras": sorted(self.req.extras or ()),
            "requires_python": str(self.requires_python),
            "editable": self.req.editable,
            "subdirectory": getattr(self.req, "subdirectory", None),
        }
        if self.req.is_vcs:
            result.update(
                {
                    self.req.vcs: self.req.repo,
                    "ref": self.req.ref,
                }
            )
            if not self.req.editable:
                result.update(revision=self.get_revision())
        elif not self.req.is_named:
            if self.req.is_file_or_url and self.req.is_local:
                self.req._root = project_root
                result.update(path=self.req.str_path)
            else:
                result.update(url=self.req.url)
        return {k: v for k, v in result.items() if v}

    def format(self) -> str:
        """Format for output."""
        return f"[req]{self.name}[/] [warning]{self.version}[/]"

    def prepare(self, environment: BaseEnvironment, reporter: CandidateReporter | None = None) -> PreparedCandidate:
        """Prepare the candidate for installation."""
        if self._prepared is None:
            if isinstance(self.req, FileRequirement):
                self.req.check_installable()
            self._prepared = PreparedCandidate(self, environment, reporter=reporter or CandidateReporter())
        else:
            self._prepared.environment = environment
            if reporter is not None:
                self._prepared.reporter = reporter
        return self._prepared


@dataclasses.dataclass
class PreparedCandidate:
    """A candidate that has been prepared for installation.
    The metadata and built wheel are available.
    """

    _build_dir_cache: ClassVar[dict[Link, str]] = {}

    candidate: Candidate
    environment: BaseEnvironment
    reporter: CandidateReporter = dataclasses.field(default_factory=CandidateReporter)

    def __post_init__(self) -> None:
        self.req = self.candidate.req
        self.link = self._replace_url_vars(self.candidate.link)

        self._cached: Path | None = None
        self._source_dir: Path | None = None
        self._unpacked_dir: Path | None = None
        self._metadata_dir: str | None = None
        self._metadata: im.Distribution | None = None

        if self.link is not None and self.link.is_file and self.link.file_path.is_dir():
            self._source_dir = self.link.file_path
            self._unpacked_dir = self._source_dir / (self.link.subdirectory or "")

    def _replace_url_vars(self, link: Link | None) -> Link | None:
        if link is None:
            return None
        url = self.environment.project.backend.expand_line(link.normalized)
        return dataclasses.replace(link, url=url)

    @cached_property
    def revision(self) -> str:
        from unearth import vcs_support

        if not (self._source_dir and os.path.exists(self._source_dir)):
            # It happens because the cached wheel is hit and the source code isn't
            # pulled to local. In this case the link url must contain the full commit
            # hash which can be taken as the revision safely.
            # See more info at https://github.com/pdm-project/pdm/issues/349
            rev = get_rev_from_url(self.candidate.link.url)  # type: ignore[union-attr]
            if rev:
                return rev
        assert isinstance(self.req, VcsRequirement)
        return vcs_support.get_backend(self.req.vcs, self.environment.project.core.ui.verbosity).get_revision(
            cast(Path, self._source_dir)
        )

    def direct_url(self) -> dict[str, Any] | None:
        """PEP 610 direct_url.json data"""
        req = self.req
        if isinstance(req, VcsRequirement):
            if req.editable:
                assert self._source_dir
                return _filter_none(
                    {
                        "url": self._source_dir.as_uri(),
                        "dir_info": {"editable": True},
                        "subdirectory": req.subdirectory,
                    }
                )
            return _filter_none(
                {
                    "url": url_without_fragments(req.repo),
                    "vcs_info": _filter_none(
                        {
                            "vcs": req.vcs,
                            "requested_revision": req.ref,
                            "commit_id": self.revision,
                        }
                    ),
                    "subdirectory": req.subdirectory,
                }
            )
        elif isinstance(req, FileRequirement):
            assert self.link is not None
            if self.link.is_file and self.link.file_path.is_dir():
                return _filter_none(
                    {
                        "url": self.link.url_without_fragment,
                        "dir_info": _filter_none({"editable": req.editable or None}),
                        "subdirectory": req.subdirectory,
                    }
                )
            hash_cache = self.environment.project.make_hash_cache()
            return _filter_none(
                {
                    "url": self.link.url_without_fragment,
                    "archive_info": {
                        "hash": hash_cache.get_hash(self.link, self.environment.session).replace(":", "=")
                    },
                    "subdirectory": req.subdirectory,
                }
            )
        else:
            return None

    def build(self) -> Path:
        """Call PEP 517 build hook to build the candidate into a wheel"""
        self._obtain(allow_all=False)
        if self._cached:
            return self._cached
        if not self.req.editable:
            cached = self._get_build_cache()
            if cached:
                return cached
        assert self._source_dir, "Source directory isn't ready yet"
        builder_cls = EditableBuilder if self.req.editable else WheelBuilder
        builder = builder_cls(str(self._unpacked_dir), self.environment)
        build_dir = self._get_wheel_dir()
        os.makedirs(build_dir, exist_ok=True)
        termui.logger.info("Running PEP 517 backend to build a wheel for %s", self.link)
        self.reporter.report_build_start(self.link.filename)  # type: ignore[union-attr]
        self._cached = Path(builder.build(build_dir, metadata_directory=self._metadata_dir))
        self.reporter.report_build_end(self.link.filename)  # type: ignore[union-attr]
        return self._cached

    def _obtain(self, allow_all: bool = False, unpack: bool = True) -> None:
        """Fetch the link of the candidate and unpack to local if necessary.

        :param allow_all: If true, don't validate the wheel tag nor hashes
        :param unpack: Whether to download and unpack the link if it's not local
        """
        if self._cached and self._wheel_compatible(self._cached.name, allow_all):
            return

        if self._source_dir and self._source_dir.exists():
            return

        sources = filtered_sources(self.environment.project.sources, self.req.key)
        env_spec = self.environment.allow_all_spec if allow_all else self.environment.spec
        with self.environment.get_finder(sources, env_spec=env_spec) as finder:
            if not self.link or (self.link.is_wheel and not self._wheel_compatible(self.link.filename, allow_all)):
                if self.req.is_file_or_url:
                    raise CandidateNotFound(f"The URL requirement {self.req.as_line()} is a wheel but incompatible")
                self.link = self._cached = None  # reset the incompatible wheel
                self.link = _find_best_match_link(
                    finder, self.req.as_pinned_version(self.candidate.version), self.candidate.hashes
                )
                if not self.link:
                    raise CandidateNotFound(
                        f"No candidate is found for `{self.req.project_name}` that matches the environment or hashes"
                    )
                if not self.candidate.link:
                    self.candidate.link = self.link
        # find if there is any build cache for the candidate
        if not self.req.editable:
            cached = self._get_build_cache()
            if cached and self._wheel_compatible(cached.name, allow_all):
                self._cached = cached
                return
        # If not, download and unpack the link
        if unpack:
            self._unpack(validate_hashes=not allow_all)

    def _unpack(self, validate_hashes: bool = False) -> None:
        hash_options = None
        if validate_hashes and self.candidate.hashes:
            hash_options = convert_hashes(self.candidate.hashes)
        assert self.link is not None
        with self.environment.get_finder() as finder:
            with TemporaryDirectory(prefix="pdm-download-") as tmpdir:
                build_dir = self._get_build_dir()
                if self.link.is_wheel:
                    download_dir = build_dir
                else:
                    download_dir = tmpdir
                result = finder.download_and_unpack(
                    self.link,
                    build_dir,
                    download_dir,
                    hash_options,
                    download_reporter=self.reporter.report_download,
                    unpack_reporter=self.reporter.report_unpack,
                )
        if self.link.is_wheel:
            self._cached = result
        else:
            self._source_dir = Path(build_dir)
            self._unpacked_dir = result

    def prepare_metadata(self, force_build: bool = False) -> im.Distribution:
        if self.candidate.installed is not None:
            return self.candidate.installed

        self._obtain(allow_all=True, unpack=False)
        if self._metadata_dir:
            return im.PathDistribution(Path(self._metadata_dir))

        if self._cached:
            return self._get_metadata_from_wheel(self._cached)

        assert self.link is not None
        if self.link.dist_info_metadata:
            assert self.link.dist_info_link
            dist = self._get_metadata_from_metadata_link(self.link.dist_info_link, self.link.dist_info_metadata)
            if dist is not None:
                return dist

        self._unpack(validate_hashes=False)
        if self._cached:  # check again if the wheel is downloaded to local
            return self._get_metadata_from_wheel(self._cached)

        assert self._unpacked_dir, "Source directory isn't ready yet"
        pyproject_toml = self._unpacked_dir / "pyproject.toml"
        if not force_build and pyproject_toml.exists():
            dist = self._get_metadata_from_project(pyproject_toml)
            if dist is not None:
                return dist

        # If all fail, try building the source to get the metadata
        metadata_parent = self.environment.project.core.create_temp_dir(prefix="pdm-meta-")
        return self._get_metadata_from_build(self._unpacked_dir, metadata_parent)

    def _get_metadata_from_metadata_link(
        self, link: Link, medata_hash: bool | dict[str, str] | None
    ) -> im.Distribution | None:
        resp = self.environment.session.get(link.normalized)
        if isinstance(medata_hash, dict):
            hash_name, hash_value = next(iter(medata_hash.items()))
            if hashlib.new(hash_name, resp.content).hexdigest() != hash_value:
                termui.logger.warning("Metadata hash mismatch for %s, ignoring the metadata", link)
                return None
        return MetadataDistribution(resp.text)

    def _get_metadata_from_wheel(self, wheel: Path) -> im.Distribution:
        # Get metadata from METADATA inside the wheel
        metadata_parent = self.environment.project.core.create_temp_dir(prefix="pdm-meta-")
        dist_info = self._metadata_dir = _get_wheel_metadata_from_wheel(wheel, metadata_parent)
        return im.PathDistribution(Path(dist_info))

    def _get_metadata_from_project(self, pyproject_toml: Path) -> im.Distribution | None:
        # Try getting from PEP 621 metadata
        from pdm.formats import MetaConvertError
        from pdm.project.project_file import PyProject

        try:
            pyproject = PyProject(pyproject_toml, ui=self.environment.project.core.ui)
        except MetaConvertError as e:
            termui.logger.warning("Failed to parse pyproject.toml: %s", e)
            return None
        metadata = pyproject.metadata
        if not metadata:
            termui.logger.warning("Failed to parse pyproject.toml")
            return None

        dynamic_fields = metadata.get("dynamic", [])
        # Use the parse result only when all are static
        if not set(dynamic_fields).isdisjoint(
            {
                "name",
                "version",
                "dependencies",
                "optional-dependencies",
                "requires-python",
            }
        ):
            return None

        try:
            backend_cls = get_backend_by_spec(pyproject.build_system)
        except Exception:
            # no variable expansion
            backend_cls = get_backend("setuptools")
        backend = backend_cls(pyproject_toml.parent)
        if "name" not in metadata:
            termui.logger.warning("Failed to parse pyproject.toml, name is required")
            return None
        setup = Setup(
            name=metadata.get("name"),
            summary=metadata.get("description"),
            version=metadata.get("version", "0.0.0"),
            install_requires=list(
                map(
                    backend.expand_line,
                    metadata.get("dependencies", []),
                )
            ),
            extras_require={
                k: list(map(backend.expand_line, v)) for k, v in metadata.get("optional-dependencies", {}).items()
            },
            python_requires=metadata.get("requires-python"),
        )
        return setup.as_dist()

    def _get_metadata_from_build(self, source_dir: Path, metadata_parent: str) -> im.Distribution:
        builder = EditableBuilder if self.req.editable else WheelBuilder
        try:
            termui.logger.info("Running PEP 517 backend to get metadata for %s", self.link)
            self.reporter.report_build_start(self.link.filename)  # type: ignore[union-attr]
            self._metadata_dir = builder(source_dir, self.environment).prepare_metadata(metadata_parent)
            self.reporter.report_build_end(self.link.filename)  # type: ignore[union-attr]
        except BuildError:
            termui.logger.warning("Failed to build package, try parsing project files.")
            try:
                setup = Setup.from_directory(source_dir)
            except Exception:
                message = "Failed to parse the project files, dependencies may be missing"
                termui.logger.warning(message)
                warnings.warn(message, PDMWarning, stacklevel=1)
                setup = Setup()
            return setup.as_dist()
        else:
            return im.PathDistribution(Path(cast(str, self._metadata_dir)))

    @property
    def metadata(self) -> im.Distribution:
        if self._metadata is None:
            result = self.prepare_metadata()
            if not self.candidate.name:
                self.req.name = self.candidate.name = cast(str, result.metadata.get("Name"))
            if not self.candidate.version and result.metadata.get("Version"):
                self.candidate.version = result.version
            if not self.candidate.requires_python:
                self.candidate.requires_python = result.metadata.get("Requires-Python", "")
            self._metadata = result
        return self._metadata

    def get_dependencies_from_metadata(self) -> list[Requirement]:
        """Get the dependencies of a candidate from metadata."""
        extras = self.req.extras or ()
        return filter_requirements_with_extras(self.metadata.requires or [], extras)

    def should_cache(self) -> bool:
        """Determine whether to cache the dependencies and built wheel."""
        from unearth import vcs_support

        if not self.environment.project.core.state.enable_cache:
            return False

        link, source_dir = self.candidate.link, self._source_dir
        if self.req.editable:
            return False
        if self.req.is_named:
            return True
        if self.req.is_vcs:
            if not source_dir:
                # If the candidate isn't prepared, we can't cache it
                return False
            assert link
            vcs_backend = vcs_support.get_backend(link.vcs, self.environment.project.core.ui.verbosity)
            return vcs_backend.is_immutable_revision(source_dir, link)
        if link and not (link.is_file and link.file_path.is_dir()):
            # Cache if the link contains egg-info like 'foo-1.0'
            return _egg_info_re.search(link.filename) is not None
        return False

    def _get_build_cache(self) -> Path | None:
        if not self.environment.project.core.state.enable_cache:
            return None
        wheel_cache = self.environment.project.make_wheel_cache()
        assert self.candidate.link
        cache_entry = wheel_cache.get(self.candidate.link, self.candidate.name, self.environment.spec)
        if cache_entry is not None:
            termui.logger.info("Using cached wheel: %s", cache_entry)
        return cache_entry

    def _get_build_dir(self) -> str:
        assert self.link is not None
        if self.link.is_file and self.link.file_path.is_dir():
            # Local directories are built in tree
            return str(self.link.file_path)
        if self.req.editable:
            # In this branch the requirement must be an editable VCS requirement.
            # The repository will be unpacked into a *persistent* src directory.
            prefix: Path | None = None
            if self.environment.is_local:
                prefix = self.environment.packages_path  # type: ignore[attr-defined]
            else:
                venv = self.environment.interpreter.get_venv()
                if venv is not None:
                    prefix = venv.root
            if prefix is not None:
                src_dir = prefix / "src"
            else:
                src_dir = Path("src")
            src_dir.mkdir(exist_ok=True, parents=True)
            dirname = self.candidate.name or self.req.name
            if not dirname:
                dirname, _ = os.path.splitext(self.link.filename)
            return str(src_dir / str(dirname))
        # Otherwise, for source dists, they will be unpacked into a *temp* directory.
        if (build_dir := self._build_dir_cache.get(self.link)) is None:
            build_dir = self._build_dir_cache[self.link] = self.environment.project.core.create_temp_dir(
                prefix="pdm-build-"
            )
        return build_dir

    def _wheel_compatible(self, wheel_file: str, allow_all: bool = False) -> bool:
        env_spec = self.environment.allow_all_spec if allow_all else self.environment.spec
        return env_spec.wheel_compatibility(wheel_file) is not None

    def _get_wheel_dir(self) -> str:
        assert self.candidate.link
        wheel_cache = self.environment.project.make_wheel_cache()
        if self.should_cache():
            termui.logger.info("Saving wheel to cache: %s", self.candidate.link)
            return wheel_cache.get_path_for_link(self.candidate.link, self.environment.spec).as_posix()
        else:
            return wheel_cache.get_ephemeral_path_for_link(self.candidate.link, self.environment.spec).as_posix()



from __future__ import annotations

from typing import TYPE_CHECKING

from pdm.cli.commands.base import BaseCommand
from pdm.cli.commands.venv.activate import ActivateCommand
from pdm.cli.commands.venv.create import CreateCommand
from pdm.cli.commands.venv.list import ListCommand
from pdm.cli.commands.venv.purge import PurgeCommand
from pdm.cli.commands.venv.remove import RemoveCommand
from pdm.cli.commands.venv.utils import get_venv_with_name
from pdm.cli.options import project_option

if TYPE_CHECKING:
    from argparse import ArgumentParser, Namespace

    from pdm.project import Project


class Command(BaseCommand):
    """Virtualenv management"""

    name = "venv"
    arguments = (project_option,)

    def add_arguments(self, parser: ArgumentParser) -> None:
        group = parser.add_mutually_exclusive_group()
        group.add_argument("--path", help="Show the path to the given virtualenv")
        group.add_argument("--python", help="Show the python interpreter path for the given virtualenv")
        subparser = parser.add_subparsers(title="commands", metavar="")
        CreateCommand.register_to(subparser, "create")
        ListCommand.register_to(subparser, "list")
        RemoveCommand.register_to(subparser, "remove")
        ActivateCommand.register_to(subparser, "activate")
        PurgeCommand.register_to(subparser, "purge")
        self.parser = parser

    def handle(self, project: Project, options: Namespace) -> None:
        if options.path:
            venv = get_venv_with_name(project, options.path)
            project.core.ui.echo(str(venv.root))
        elif options.python:
            venv = get_venv_with_name(project, options.python)
            project.core.ui.echo(str(venv.interpreter))
        else:
            self.parser.print_help()



from __future__ import annotations

import argparse
import os
from typing import TYPE_CHECKING

from pdm.cli.commands import build
from pdm.cli.commands.base import BaseCommand
from pdm.cli.commands.publish.package import PackageFile
from pdm.cli.hooks import HookManager
from pdm.cli.options import project_option, skip_option, verbose_option
from pdm.exceptions import PdmUsageError, PublishError
from pdm.termui import logger

if TYPE_CHECKING:
    from httpx import Response

    from pdm.cli.commands.publish.repository import Repository
    from pdm.project import Project


class Command(BaseCommand):
    """Build and publish the project to PyPI"""

    arguments = (verbose_option, project_option, skip_option)

    def add_arguments(self, parser: argparse.ArgumentParser) -> None:
        parser.add_argument(
            "-r",
            "--repository",
            help="The repository name or url to publish the package to [env var: PDM_PUBLISH_REPO]",
        )
        parser.add_argument(
            "-u",
            "--username",
            help="The username to access the repository [env var: PDM_PUBLISH_USERNAME]",
        )
        parser.add_argument(
            "-P",
            "--password",
            help="The password to access the repository [env var: PDM_PUBLISH_PASSWORD]",
        )
        parser.add_argument(
            "-S",
            "--sign",
            action="store_true",
            help="Upload the package with PGP signature",
        )
        parser.add_argument(
            "-i",
            "--identity",
            help="GPG identity used to sign files.",
        )
        parser.add_argument(
            "-c",
            "--comment",
            help="The comment to include with the distribution file.",
        )
        parser.add_argument(
            "--no-build",
            action="store_false",
            dest="build",
            help="Don't build the package before publishing",
        )
        parser.add_argument(
            "-d",
            "--dest",
            help="The directory to upload the package from",
            default="dist",
        )
        parser.add_argument(
            "--skip-existing",
            action="store_true",
            help="Skip uploading files that already exist. This may not work with some repository implementations.",
        )
        group = parser.add_mutually_exclusive_group()
        group.add_argument(
            "--no-very-ssl", action="store_false", dest="verify_ssl", help="Disable SSL verification", default=None
        )
        group.add_argument(
            "--ca-certs",
            dest="ca_certs",
            help="The path to a PEM-encoded Certificate Authority bundle to use"
            " for publish server validation [env var: PDM_PUBLISH_CA_CERTS]",
        )

    @staticmethod
    def _make_package(filename: str, signatures: dict[str, str], options: argparse.Namespace) -> PackageFile:
        p = PackageFile.from_filename(filename, options.comment)
        if p.base_filename in signatures:
            p.add_gpg_signature(signatures[p.base_filename], p.base_filename + ".asc")
        elif options.sign:
            p.sign(options.identity)
        return p

    @staticmethod
    def _skip_upload(response: Response) -> bool:
        status = response.status_code
        reason = response.reason_phrase.lower()
        text = response.text.lower()

        # Borrowed from https://github.com/pypa/twine/blob/main/twine/commands/upload.py#L149
        return (
            # pypiserver (https://pypi.org/project/pypiserver)
            status == 409
            # PyPI / TestPyPI / GCP Artifact Registry
            or (status == 400 and any("already exist" in x for x in [reason, text]))
            # Nexus Repository OSS (https://www.sonatype.com/nexus-repository-oss)
            or (
                status == 400
                and any(token in x for x in [reason, text] for token in ["updating asset", "cannot be updated"])
            )
            # Artifactory (https://jfrog.com/artifactory/)
            or (status == 403 and "overwrite artifact" in text)
            # Gitlab Enterprise Edition (https://about.gitlab.com)
            or (status == 400 and "already been taken" in text)
        )

    @staticmethod
    def _check_response(response: Response) -> None:
        import httpx

        message = ""
        if response.status_code == 410 and "pypi.python.org" in str(response.url):
            message = (
                "Uploading to these sites is deprecated. "
                "Try using https://upload.pypi.org/legacy/ "
                "(or https://test.pypi.org/legacy/) instead."
            )
        elif response.status_code == 405 and "pypi.org" in str(response.url):
            message = "It appears you're trying to upload to pypi.org but have an invalid URL."
        else:
            try:
                response.raise_for_status()
            except httpx.HTTPStatusError as err:
                message = str(err)
                if response.text:
                    logger.debug(response.text)
        if message:
            raise PublishError(message)

    @staticmethod
    def get_repository(project: Project, options: argparse.Namespace) -> Repository:
        from pdm.cli.commands.publish.repository import Repository

        repository = options.repository or os.getenv("PDM_PUBLISH_REPO", "pypi")
        username = options.username or os.getenv("PDM_PUBLISH_USERNAME")
        password = options.password or os.getenv("PDM_PUBLISH_PASSWORD")
        ca_certs = options.ca_certs or os.getenv("PDM_PUBLISH_CA_CERTS")

        config = project.project_config.get_repository_config(repository, "repository")
        if config is None:
            config = project.global_config.get_repository_config(repository, "repository")
            if config is None:
                raise PdmUsageError(f"Missing repository config of {repository}")
        else:
            global_config = project.global_config.get_repository_config(repository, "repository")
            if global_config is not None:
                config.passive_update(global_config)

        assert config.url is not None
        if username is not None:
            config.username = username
        if password is not None:
            config.password = password
        if ca_certs is not None:
            config.ca_certs = ca_certs
        if options.verify_ssl is False:
            config.verify_ssl = options.verify_ssl
        config.populate_keyring_auth()
        return Repository(project, config)

    def handle(self, project: Project, options: argparse.Namespace) -> None:
        hooks = HookManager(project, options.skip)

        hooks.try_emit("pre_publish")

        if options.build:
            build.Command.do_build(project, dest=options.dest, hooks=hooks)

        upload_dir = project.root.joinpath(options.dest)
        package_files = [str(p) for p in upload_dir.iterdir() if not p.name.endswith(".asc")]
        signatures = {p.stem: str(p) for p in upload_dir.iterdir() if p.name.endswith(".asc")}

        repository = self.get_repository(project, options)
        uploaded: list[PackageFile] = []
        with project.core.ui.logging("publish"):
            packages = sorted(
                (self._make_package(p, signatures, options) for p in package_files),
                # Upload wheels first if they exist.
                key=lambda p: not p.base_filename.endswith(".whl"),
            )
            for package in packages:
                resp = repository.upload(package)
                logger.debug("Response from %s:\n%s %s", resp.url, resp.status_code, resp.reason_phrase)

                if options.skip_existing and self._skip_upload(resp):
                    project.core.ui.warn(f"Skipping {package.base_filename} because it appears to already exist")
                    continue
                self._check_response(resp)
                uploaded.append(package)

        release_urls = repository.get_release_urls(uploaded)
        if release_urls:
            project.core.ui.echo("\n[success]View at:")
            for url in release_urls:
                project.core.ui.echo(url)

        hooks.try_emit("post_publish")



from __future__ import annotations

import contextlib
import enum
import logging
import os
import tempfile
import warnings
from typing import IO, TYPE_CHECKING

import rich
from rich.box import ROUNDED
from rich.console import Console
from rich.progress import Progress, ProgressColumn
from rich.prompt import Confirm, IntPrompt, Prompt
from rich.table import Table
from rich.theme import Theme

from pdm.exceptions import PDMWarning

if TYPE_CHECKING:
    from typing import Any, Iterator, Sequence

    from pdm._types import RichProtocol, Spinner, SpinnerT

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())
logger.propagate = False
logger.setLevel(logging.DEBUG)
unearth_logger = logging.getLogger("unearth")
unearth_logger.addHandler(logging.NullHandler())
unearth_logger.propagate = False
unearth_logger.setLevel(logging.DEBUG)

DEFAULT_THEME = {
    "primary": "cyan",
    "success": "green",
    "warning": "yellow",
    "error": "red",
    "info": "blue",
    "req": "bold green",
}
rich.reconfigure(highlight=False, theme=Theme(DEFAULT_THEME))
_err_console = Console(stderr=True, theme=Theme(DEFAULT_THEME))


def is_interactive(console: Console | None = None) -> bool:
    """Check if the terminal is run under interactive mode"""
    if console is None:
        console = rich.get_console()
    return "PDM_NON_INTERACTIVE" not in os.environ and console.is_interactive


def is_legacy_windows(console: Console | None = None) -> bool:
    """Legacy Windows renderer may have problem rendering emojis"""
    if console is None:
        console = rich.get_console()
    return console.legacy_windows


def style(text: str, *args: str, style: str | None = None, **kwargs: Any) -> str:
    """return text with ansi codes using rich console

    :param text: message with rich markup, defaults to "".
    :param style: rich style to apply to whole string
    :return: string containing ansi codes
    """
    _console = rich.get_console()
    if _console.legacy_windows or not _console.is_terminal:  # pragma: no cover
        return text
    with _console.capture() as capture:
        _console.print(text, *args, end="", style=style, **kwargs)
    return capture.get()


def confirm(*args: str, **kwargs: Any) -> bool:
    default = kwargs.setdefault("default", False)
    if not is_interactive():
        return default
    return Confirm.ask(*args, **kwargs)


def ask(*args: str, prompt_type: type[str] | type[int] | None = None, **kwargs: Any) -> str:
    """prompt user and return response

    :prompt_type: which rich prompt to use, defaults to str.
    :raises ValueError: unsupported prompt type
    :return: str of user's selection
    """
    if not prompt_type or prompt_type is str:
        return Prompt.ask(*args, **kwargs)
    elif prompt_type is int:
        return str(IntPrompt.ask(*args, **kwargs))
    else:
        raise ValueError(f"unsupported {prompt_type}")


class Verbosity(enum.IntEnum):
    QUIET = -1
    NORMAL = 0
    DETAIL = 1
    DEBUG = 2


LOG_LEVELS = {
    Verbosity.NORMAL: logging.WARN,
    Verbosity.DETAIL: logging.INFO,
    Verbosity.DEBUG: logging.DEBUG,
}


class Emoji:
    if is_legacy_windows():
        SUCC = "v"
        FAIL = "x"
        LOCK = " "
        POPPER = " "
        ELLIPSIS = "..."
        ARROW_SEPARATOR = ">"
    else:
        SUCC = ":heavy_check_mark:"
        FAIL = ":heavy_multiplication_x:"
        LOCK = ":lock:"
        POPPER = ":party_popper:"
        ELLIPSIS = "…"
        ARROW_SEPARATOR = "➤"


if is_legacy_windows():
    SPINNER = "line"
else:
    SPINNER = "dots"


class DummySpinner:
    """A dummy spinner class implementing needed interfaces.
    But only display text onto screen.
    """

    def __init__(self, text: str) -> None:
        self.text = text

    def _show(self) -> None:
        _err_console.print(f"[primary]STATUS:[/] {self.text}")

    def update(self, text: str) -> None:
        self.text = text
        self._show()

    def __enter__(self: SpinnerT) -> SpinnerT:
        self._show()  # type: ignore[attr-defined]
        return self

    def __exit__(self, *args: Any) -> None:
        pass


class TruncatedIO:
    """A wrapper for IO that truncates output after certain length."""

    def __init__(self, wrapped: IO[str], max_length: int = 100 * 1024 * 1024) -> None:
        self.max_length = max_length
        self._fp = wrapped
        self._truncated = False

    def write(self, s: str) -> int:
        if self._truncated:
            return 0
        if self._fp.tell() >= self.max_length:
            s = "...[truncated]...\n"
            self._truncated = True
            _err_console.print("[warning]WARNING:[/] Log truncated due to excessive length.")
        return self._fp.write(s)

    def __getattr__(self, name: str) -> Any:
        return getattr(self._fp, name)


class UI:
    """Terminal UI object"""

    MAX_LOG_SIZE = 100 * 1024 * 1024  # 100MB

    def __init__(
        self, verbosity: Verbosity = Verbosity.NORMAL, *, exit_stack: contextlib.ExitStack | None = None
    ) -> None:
        self.verbosity = verbosity
        self.exit_stack = exit_stack or contextlib.ExitStack()
        self.log_dir: str | None = None

    def set_verbosity(self, verbosity: int) -> None:
        self.verbosity = Verbosity(verbosity)
        if self.verbosity == Verbosity.QUIET:
            self.exit_stack.enter_context(warnings.catch_warnings())
            warnings.simplefilter("ignore", PDMWarning, append=True)
            warnings.simplefilter("ignore", FutureWarning, append=True)

    def set_theme(self, theme: Theme) -> None:
        """set theme for rich console

        :param theme: dict of theme
        """
        rich.get_console().push_theme(theme)
        _err_console.push_theme(theme)

    def echo(
        self,
        message: str | RichProtocol = "",
        err: bool = False,
        verbosity: Verbosity = Verbosity.QUIET,
        **kwargs: Any,
    ) -> None:
        """print message using rich console

        :param message: message with rich markup, defaults to "".
        :param err: if true print to stderr, defaults to False.
        :param verbosity: verbosity level, defaults to QUIET.
        """
        if self.verbosity >= verbosity:
            console = _err_console if err else rich.get_console()
            if not console.is_interactive:
                kwargs.setdefault("crop", False)
                kwargs.setdefault("overflow", "ignore")
            console.print(message, **kwargs)

    def display_columns(self, rows: Sequence[Sequence[str]], header: list[str] | None = None) -> None:
        """Print rows in aligned columns.

        :param rows: a rows of data to be displayed.
        :param header: a list of header strings.
        """

        if header:
            table = Table(box=ROUNDED)
            for title in header:
                if title[0] == "^":
                    title, justify = title[1:], "center"
                elif title[0] == ">":
                    title, justify = title[1:], "right"
                else:
                    title, justify = title, "left"
                table.add_column(title, justify=justify)
        else:
            table = Table.grid(padding=(0, 1))
            for _ in rows[0]:
                table.add_column()
        for row in rows:
            table.add_row(*row)

        rich.print(table)

    @contextlib.contextmanager
    def logging(self, type_: str = "install") -> Iterator[logging.Logger]:
        """A context manager that opens a file for logging when verbosity is NORMAL or
        print to the stdout otherwise.
        """
        log_file: str | None = None
        if self.verbosity >= Verbosity.DETAIL:
            handler: logging.Handler = logging.StreamHandler()
            handler.setLevel(LOG_LEVELS[self.verbosity])
        else:
            if self.log_dir and not os.path.exists(self.log_dir):
                os.makedirs(self.log_dir, exist_ok=True)
            self._clean_logs()
            fp, log_file = tempfile.mkstemp(".log", f"pdm-{type_}-", self.log_dir)
            handler = logging.StreamHandler(
                TruncatedIO(self.exit_stack.enter_context(open(fp, "a", encoding="utf-8")), self.MAX_LOG_SIZE)
            )
            handler.setLevel(logging.DEBUG)

        handler.setFormatter(logging.Formatter("%(name)s: %(message)s"))
        logger.addHandler(handler)
        unearth_logger.addHandler(handler)

        def cleanup() -> None:
            if not log_file:
                return
            with contextlib.suppress(OSError):
                os.unlink(log_file)

        try:
            yield logger
        except Exception:
            if self.verbosity < Verbosity.DETAIL:
                logger.exception("Error occurs")
                self.echo(
                    f"See [warning]{log_file}[/] for detailed debug log.",
                    style="error",
                    err=True,
                )
            raise
        else:
            self.exit_stack.callback(cleanup)
        finally:
            logger.removeHandler(handler)
            unearth_logger.removeHandler(handler)
            handler.close()

    def open_spinner(self, title: str) -> Spinner:
        """Open a spinner as a context manager."""
        if self.verbosity >= Verbosity.DETAIL or not is_interactive():
            return DummySpinner(title)
        else:
            return _err_console.status(title, spinner=SPINNER, spinner_style="primary")

    def make_progress(self, *columns: str | ProgressColumn, **kwargs: Any) -> Progress:
        """create a progress instance for indented spinners"""
        return Progress(*columns, disable=self.verbosity >= Verbosity.DETAIL, **kwargs)

    def info(self, message: str, verbosity: Verbosity = Verbosity.NORMAL) -> None:
        """Print a message to stdout."""
        self.echo(f"[info]INFO:[/] [dim]{message}[/]", err=True, verbosity=verbosity)

    def deprecated(self, message: str, verbosity: Verbosity = Verbosity.NORMAL) -> None:
        """Print a message to stdout."""
        self.echo(f"[warning]DEPRECATED:[/] [dim]{message}[/]", err=True, verbosity=verbosity)

    def warn(self, message: str, verbosity: Verbosity = Verbosity.NORMAL) -> None:
        """Print a message to stdout."""
        self.echo(f"[warning]WARNING:[/] {message}", err=True, verbosity=verbosity)

    def error(self, message: str, verbosity: Verbosity = Verbosity.QUIET) -> None:
        """Print a message to stdout."""
        self.echo(f"[error]ERROR:[/] {message}", err=True, verbosity=verbosity)

    def _clean_logs(self) -> None:
        import time
        from pathlib import Path

        if self.log_dir is None:
            return
        for file in Path(self.log_dir).iterdir():
            if not file.is_file():
                continue
            if file.stat().st_ctime < time.time() - 7 * 24 * 60 * 60:  # 7 days
                file.unlink()



from __future__ import annotations

import os
from typing import TYPE_CHECKING, Any, Iterable, cast
from urllib.parse import urlparse, urlunparse

import httpx
from id import AmbientCredentialError, detect_credential
from rich.progress import (
    BarColumn,
    DownloadColumn,
    TimeRemainingColumn,
    TransferSpeedColumn,
)

from pdm import termui
from pdm.cli.commands.publish.package import PackageFile
from pdm.exceptions import PdmUsageError
from pdm.project import Project
from pdm.project.config import DEFAULT_REPOSITORIES

if TYPE_CHECKING:
    from typing import Callable, Self

    from httpx._multipart import MultipartStream

    from pdm._types import RepositoryConfig


class CallbackWrapperStream(httpx.SyncByteStream):
    def __init__(self, stream: httpx.SyncByteStream, callback: Callable[[Self], Any]) -> None:
        self._stream = stream
        self._callback = callback
        self.bytes_read = 0

    def __iter__(self) -> Iterable[bytes]:
        for chunk in self._stream:
            self.bytes_read += len(chunk)
            self._callback(self)
            yield chunk


class Repository:
    def __init__(self, project: Project, config: RepositoryConfig) -> None:
        self.url = cast(str, config.url)
        self.session = project.environment._build_session([config])

        self._credentials_to_save: tuple[str, str, str] | None = None
        self.ui = project.core.ui
        username, password = self._ensure_credentials(config.username, config.password)
        self.session.auth = (username, password)

    def _ensure_credentials(self, username: str | None, password: str | None) -> tuple[str, str]:
        from pdm.models.auth import keyring

        parsed_url = urlparse(self.url)
        netloc = parsed_url.netloc
        if username and password:
            return username, password
        if password:
            return "__token__", password
        if parsed_url.username is not None and parsed_url.password is not None:
            return parsed_url.username, parsed_url.password
        if keyring.enabled:
            auth = keyring.get_auth_info(self.url, username)
            if auth is not None:
                return auth
        token = self._get_pypi_token_via_oidc()
        if token is not None:
            return "__token__", token
        if not termui.is_interactive():
            raise PdmUsageError("Username and password are required")
        username, password, save = self._prompt_for_credentials(netloc, username)
        if save and keyring.enabled and termui.confirm("Save credentials to keyring?"):
            self._credentials_to_save = (netloc, username, password)
        return username, password

    def _get_pypi_token_via_oidc(self) -> str | None:
        self.ui.echo("Getting PyPI token via OIDC...")

        try:
            parsed_url = urlparse(self.url)
            audience_url = urlunparse(parsed_url._replace(path="/_/oidc/audience"))
            resp = self.session.get(audience_url)
            resp.raise_for_status()
            audience = cast(str, resp.json()["audience"])
            oidc_token = detect_credential(audience)
            if oidc_token is None:
                self.ui.echo(
                    "This platform is not supported for trusted publishing via OIDC",
                    err=True,
                )
                return None
            mint_token_url = urlunparse(parsed_url._replace(path="/_/oidc/mint-token"))
            resp = self.session.post(mint_token_url, json={"token": oidc_token})
            resp.raise_for_status()
            token = resp.json()["token"]
        except AmbientCredentialError as e:
            self.ui.echo(f"Unable to detect OIDC token for CI platform: {e}", err=True)
            return None
        except httpx.HTTPError:
            self.ui.echo("Failed to get PyPI token via OIDC", err=True)
            return None
        else:
            if os.getenv("GITHUB_ACTIONS"):
                # tell GitHub Actions to mask the token in any console logs
                print(f"::add-mask::{token}")
            return token

    def _prompt_for_credentials(self, service: str, username: str | None) -> tuple[str, str, bool]:
        from pdm.models.auth import keyring

        if keyring.enabled:
            cred = keyring.get_auth_info(service, username)
            if cred is not None:
                return cred[0], cred[1], False
        if username is None:
            username = termui.ask("[primary]Username")
        password = termui.ask("[primary]Password", password=True)
        return username, password, True

    def _save_credentials(self, service: str, username: str, password: str) -> None:
        from pdm.models.auth import keyring

        self.ui.echo("Saving credentials to keyring")
        keyring.save_auth_info(service, username, password)

    def get_release_urls(self, packages: list[PackageFile]) -> Iterable[str]:
        if self.url.startswith(DEFAULT_REPOSITORIES["pypi"].rstrip("/")):
            base = "https://pypi.org/"
        elif self.url.startswith(DEFAULT_REPOSITORIES["testpypi"].rstrip("/")):
            base = "https://test.pypi.org/"
        else:
            return set()
        return {f"{base}project/{package.metadata['name']}/{package.metadata['version']}/" for package in packages}

    def upload(self, package: PackageFile) -> httpx.Response:
        data_fields = package.metadata_dict
        data_fields.update(
            {
                ":action": "file_upload",
                "protocol_version": "1",
            }
        )
        with self.ui.make_progress(
            " [progress.percentage]{task.percentage:>3.0f}%",
            BarColumn(),
            DownloadColumn(),
            "•",
            TimeRemainingColumn(
                compact=True,
                elapsed_when_finished=True,
            ),
            "•",
            TransferSpeedColumn(),
        ) as progress:
            progress.console.print(f"Uploading [success]{package.base_filename}")

            with open(package.filename, "rb") as fp:
                file_fields = [("content", (package.base_filename, fp, "application/octet-stream"))]

                def on_upload(monitor: CallbackWrapperStream) -> None:
                    progress.update(job, completed=monitor.bytes_read)

                request = self.session.build_request("POST", self.url, data=data_fields, files=file_fields)
                stream = cast("MultipartStream", request.stream)
                request.stream = CallbackWrapperStream(stream, on_upload)

                job = progress.add_task("", total=stream.get_content_length())
                resp = self.session.send(request, follow_redirects=False)
                if not resp.is_error and self._credentials_to_save is not None:
                    self._save_credentials(*self._credentials_to_save)
                    self._credentials_to_save = None
                return resp



from __future__ import annotations

import dataclasses
import os
from collections import defaultdict
from functools import cached_property
from typing import TYPE_CHECKING, Callable

from packaging.specifiers import InvalidSpecifier
from packaging.version import InvalidVersion
from resolvelib import AbstractProvider, RequirementsConflicted
from resolvelib.resolvers import Criterion

from pdm.exceptions import CandidateNotFound, InvalidPyVersion, RequirementError
from pdm.models.candidates import Candidate
from pdm.models.repositories import LockedRepository
from pdm.models.requirements import FileRequirement, Requirement, VcsRequirement, parse_requirement, strip_extras
from pdm.models.specifiers import PySpecSet
from pdm.resolver.python import PythonCandidate, PythonRequirement, find_python_matches, is_python_satisfied_by
from pdm.termui import logger
from pdm.utils import (
    deprecation_warning,
    get_requirement_from_override,
    normalize_name,
    parse_version,
    url_without_fragments,
)

if TYPE_CHECKING:
    from typing import Iterable, Iterator, Mapping, Sequence, TypeVar

    from resolvelib.resolvers import RequirementInformation

    from pdm._types import Comparable
    from pdm.models.repositories import BaseRepository, LockedRepository
    from pdm.models.requirements import Requirement
    from pdm.models.working_set import WorkingSet

    ProviderT = TypeVar("ProviderT", bound="type[BaseProvider]")


_PROVIDER_REGISTRY: dict[str, type[BaseProvider]] = {}
_CONFLICT_PRIORITY_THRESHOLD = 5


def get_provider(strategy: str) -> type[BaseProvider]:
    return _PROVIDER_REGISTRY[strategy]


def register_provider(strategy: str) -> Callable[[ProviderT], ProviderT]:
    def wrapper(cls: ProviderT) -> ProviderT:
        _PROVIDER_REGISTRY[strategy] = cls
        return cls

    return wrapper


@register_provider("all")
class BaseProvider(AbstractProvider[Requirement, Candidate, str]):
    def __init__(
        self,
        repository: BaseRepository,
        allow_prereleases: bool | None = None,
        overrides: dict[str, str] | None = None,
        direct_minimal_versions: bool = False,
        *,
        locked_repository: LockedRepository | None = None,
    ) -> None:
        if overrides is not None:  # pragma: no cover
            deprecation_warning(
                "The `overrides` argument is deprecated and will be removed in the future.", stacklevel=2
            )
        if allow_prereleases is not None:  # pragma: no cover
            deprecation_warning(
                "The `allow_prereleases` argument is deprecated and will be removed in the future.", stacklevel=2
            )
        project = repository.environment.project
        self.repository = repository
        self.allow_prereleases = project.pyproject.allow_prereleases  # Root allow_prereleases value
        self.fetched_dependencies: dict[tuple[str, str | None], list[Requirement]] = {}
        self.excludes = {normalize_name(k) for k in project.pyproject.resolution.get("excludes", [])}
        self.direct_minimal_versions = direct_minimal_versions
        self.locked_repository = locked_repository
        self._conflict_counts: defaultdict[str, int] = defaultdict(int)
        self._conflict_promoted: set[str] = set()

    def requirement_preference(self, requirement: Requirement) -> Comparable:
        """Return the preference of a requirement to find candidates.

        - Editable requirements are preferred.
        - File links are preferred.
        - The one with narrower specifierset is preferred.
        """
        editable = requirement.editable
        is_named = requirement.is_named
        is_pinned = requirement.is_pinned
        is_prerelease = bool(requirement.prerelease) or bool(requirement.specifier.prereleases)
        specifier_parts = len(requirement.specifier)
        return (not editable, is_named, not is_pinned, not is_prerelease, -specifier_parts)

    def identify(self, requirement_or_candidate: Requirement | Candidate) -> str:
        return requirement_or_candidate.identify()

    def narrow_requirement_selection(
        self,
        identifiers: Iterable[str],
        resolutions: Mapping[str, Candidate],
        candidates: Mapping[str, Iterator[Candidate]],
        information: Mapping[str, Iterator[RequirementInformation]],
        backtrack_causes: Sequence[RequirementInformation],
    ) -> Iterable[str]:
        backtrack_identifiers: set[str] = set()
        for requirement, parent in backtrack_causes:
            names = [requirement.identify()]
            if parent is not None:
                names.append(parent.identify())
            for name in names:
                backtrack_identifiers.add(name)
                if name not in resolutions:
                    self._conflict_counts[name] += 1
                    if self._conflict_counts[name] >= _CONFLICT_PRIORITY_THRESHOLD:
                        self._conflict_promoted.add(name)

        current_backtrack_causes: list[str] = []
        promoted: list[str] = []
        for identifier in identifiers:
            if identifier == "python":
                return [identifier]
            if identifier in backtrack_identifiers:
                current_backtrack_causes.append(identifier)
                continue
            if identifier in self._conflict_promoted:
                promoted.append(identifier)

        if current_backtrack_causes:
            return current_backtrack_causes
        if promoted:
            return promoted
        return identifiers

    def get_preference(
        self,
        identifier: str,
        resolutions: Mapping[str, Candidate],
        candidates: Mapping[str, Iterator[Candidate]],
        information: Mapping[str, Iterator[RequirementInformation]],
        backtrack_causes: Sequence[RequirementInformation],
    ) -> tuple[Comparable, ...]:
        is_top = any(parent is None for _, parent in information[identifier])
        backtrack_identifiers = {req.identify() for req, _ in backtrack_causes} | {
            parent.identify() for _, parent in backtrack_causes if parent is not None
        }
        # Use the REAL identifier as it may be updated after candidate preparation.
        deps: list[Requirement] = []
        for candidate in candidates[identifier]:
            try:
                deps = self.get_dependencies(candidate)
            except RequirementsConflicted:
                continue
            break
        is_backtrack_cause = any(dep.identify() in backtrack_identifiers for dep in deps)
        is_file_or_url = any(not requirement.is_named for requirement, _ in information[identifier])
        operators = [spec.operator for req, _ in information[identifier] for spec in req.specifier]
        is_python = identifier == "python"
        is_pinned = any(op[:2] == "==" for op in operators)
        constraints = len(operators)
        is_conflict_promoted = identifier in self._conflict_promoted
        return (
            not is_python,
            not is_top,
            not is_conflict_promoted,
            not is_file_or_url,
            not is_pinned,
            not is_backtrack_cause,
            -constraints,
            identifier,
        )

    @cached_property
    def locked_candidates(self) -> dict[str, list[Candidate]]:
        return self.locked_repository.all_candidates if self.locked_repository else {}

    @cached_property
    def overrides(self) -> dict[str, Requirement]:
        """A mapping of package name to the requirement for overriding."""
        from pdm.formats.requirements import RequirementParser

        project_overrides: dict[str, str] = {
            normalize_name(k): v
            for k, v in self.repository.environment.project.pyproject.resolution.get("overrides", {}).items()
        }
        requirements: dict[str, Requirement] = {}
        for name, value in project_overrides.items():
            req = get_requirement_from_override(name, value)
            r = parse_requirement(req)
            requirements[r.identify()] = r

        # Read from --override files
        parser = RequirementParser(self.repository.environment.session)
        for override_file in self.repository.environment.project.core.state.overrides:
            parser.parse_file(override_file)
        for r in parser.requirements:
            # There might be duplicates, we only keep the last one
            requirements[r.identify()] = r

        return requirements

    def _is_direct_requirement(self, requirement: Requirement) -> bool:
        from itertools import chain

        project = self.repository.environment.project
        all_dependencies = chain.from_iterable(project.all_dependencies.values())
        return any(r.is_named and requirement.identify() == r.identify() for r in all_dependencies)

    def _find_candidates(self, requirement: Requirement) -> Iterable[Candidate]:
        if not requirement.is_named and not isinstance(self.repository, LockedRepository):
            can = Candidate(requirement)
            if not can.name:
                can.prepare(self.repository.environment).metadata
            yield can
        else:
            prerelease = requirement.prerelease
            if prerelease is None and requirement.is_pinned and requirement.specifier.prereleases:
                prerelease = True
            if prerelease is None and (key := requirement.identify()) in self.locked_candidates:
                # keep the prerelease if it is locked
                candidates = self.locked_candidates[key]
                for candidate in candidates:
                    if candidate.version is not None:
                        try:
                            parsed_version = parse_version(candidate.version)
                        except InvalidVersion:  # pragma: no cover
                            pass
                        else:
                            if parsed_version.is_prerelease:
                                prerelease = True
                                break
            found = self.repository.find_candidates(
                requirement,
                self.allow_prereleases if prerelease is None else prerelease,
                minimal_version=self.direct_minimal_versions and self._is_direct_requirement(requirement),
            )

            current_version: str | None = None
            collected_wheels: list[Candidate] = []
            collected_others: list[Candidate] = []
            for candidate in found:
                assert candidate.version is not None
                if current_version is None:
                    current_version = candidate.version
                if candidate.version != current_version:
                    # If there are wheels for the given version, we should only return wheels
                    # to avoid build steps.
                    if collected_wheels:
                        yield collected_wheels[0]
                    elif collected_others:
                        yield collected_others[0]
                    current_version = candidate.version
                    collected_wheels.clear()
                    collected_others.clear()
                if candidate.link and candidate.link.is_wheel:
                    collected_wheels.append(candidate)
                else:
                    collected_others.append(candidate)
            if collected_wheels:
                yield collected_wheels[0]
            elif collected_others:
                yield collected_others[0]

    def find_matches(
        self,
        identifier: str,
        requirements: Mapping[str, Iterator[Requirement]],
        incompatibilities: Mapping[str, Iterator[Candidate]],
    ) -> Callable[[], Iterator[Candidate]]:
        def matches_gen() -> Iterator[Candidate]:
            incompat = list(incompatibilities[identifier])
            if identifier == "python":
                candidates = find_python_matches(identifier, requirements)
                return (c for c in candidates if c not in incompat)
            elif identifier in self.overrides:
                return iter(self._find_candidates(self.overrides[identifier]))
            else:
                name, extras = strip_extras(identifier)
                if name in self.overrides:
                    req = dataclasses.replace(self.overrides[name], extras=extras)
                    return iter(self._find_candidates(req))
            reqs = list(requirements[identifier])
            if not reqs:
                return iter(())
            original_req = min(reqs, key=self.requirement_preference)
            bare_name, extras = strip_extras(identifier)
            if extras and bare_name in requirements:
                # We should consider the requirements for both foo and foo[extra]
                reqs.extend(requirements[bare_name])
            reqs.sort(key=self.requirement_preference)
            candidates = self._find_candidates(reqs[0])
            return (
                # In some cases we will use candidates from the bare requirement,
                # this will miss the extra dependencies if any. So we associate the original
                # requirement back with the candidate since it is used by `get_dependencies()`.
                can.copy_with(original_req) if extras else can
                for can in candidates
                if can not in incompat and all(self.is_satisfied_by(r, can) for r in reqs)
            )

        return matches_gen

    def _compare_file_reqs(self, req1: FileRequirement, req2: FileRequirement) -> bool:
        backend = self.repository.environment.project.backend
        if req1.path and req2.path:
            return os.path.normpath(req1.path.absolute()) == os.path.normpath(req2.path.absolute())
        left = backend.expand_line(url_without_fragments(req1.get_full_url()))
        right = backend.expand_line(url_without_fragments(req2.get_full_url()))
        return left == right

    def is_satisfied_by(self, requirement: Requirement, candidate: Candidate) -> bool:
        if isinstance(requirement, PythonRequirement):
            return is_python_satisfied_by(requirement, candidate)
        elif (name := candidate.identify()) in self.overrides or strip_extras(name)[0] in self.overrides:
            return True
        if not requirement.is_named:
            if candidate.req.is_named:
                return False
            can_req = candidate.req
            if requirement.is_vcs and can_req.is_vcs:
                return can_req.vcs == requirement.vcs and can_req.repo == requirement.repo  # type: ignore[attr-defined]
            return self._compare_file_reqs(requirement, can_req)  # type: ignore[arg-type]
        version = candidate.version
        this_name = self.repository.environment.project.name
        if version is None or candidate.name == this_name:
            # This should be a URL candidate or self package, consider it to be matching
            return True
        # Allow prereleases if: 1) it is not specified in the tool settings or
        # 2) the candidate doesn't come from PyPI index or 3) the requirement is pinned
        allow_prereleases = (
            self.allow_prereleases in (True, None) or not candidate.req.is_named or requirement.is_pinned
        )
        return requirement.specifier.contains(version, allow_prereleases)

    def _get_dependencies_from_repository(self, candidate: Candidate) -> tuple[list[Requirement], PySpecSet, str]:
        return self.repository.get_dependencies(candidate)

    def get_dependencies(self, candidate: Candidate) -> list[Requirement]:
        if isinstance(candidate, PythonCandidate):
            return []
        try:
            deps, requires_python, _ = self._get_dependencies_from_repository(candidate)
        except (RequirementError, InvalidPyVersion, InvalidSpecifier) as e:
            # When the metadata is invalid, skip this candidate by marking it as conflicting.
            # Here we pass an empty criterion so it doesn't provide any info to the resolution.
            logger.error("Invalid metadata in %s: %s", candidate, e)
            raise RequirementsConflicted(Criterion([], [], [])) from None

        if candidate.req.extras:
            # XXX: If the requirement has extras, add the original candidate
            # (without extras) as its dependency. This ensures the same package with
            # different extras resolve to the same version.
            self_req = dataclasses.replace(
                candidate.req.as_pinned_version(candidate.version),
                extras=None,
                marker=None,
            )
            if self_req not in deps:
                deps.insert(0, self_req)
        self.fetched_dependencies[candidate.dep_key] = deps[:]
        # Filter out incompatible dependencies(e.g. functools32) early so that
        # we don't get errors when building wheels.
        valid_deps: list[Requirement] = []
        for dep in deps:
            if (
                dep.requires_python
                & requires_python
                & candidate.req.requires_python
                & PySpecSet(self.repository.env_spec.requires_python)
            ).is_empty():
                continue
            if dep.marker and not dep.marker.matches(self.repository.env_spec):
                continue
            if dep.identify() in self.excludes:
                continue
            dep.requires_python &= candidate.req.requires_python
            valid_deps.append(dep)
        # A candidate contributes to the Python requirements only when:
        # It isn't an optional dependency, or the requires-python doesn't cover
        # the req's requires-python.
        # For example, A v1 requires python>=3.6, it not eligible on a project with
        # requires-python=">=2.7". But it is eligible if A has environment marker
        # A1; python_version>='3.8'
        new_requires_python = candidate.req.requires_python & self.repository.environment.python_requires
        if not (
            candidate.identify() in self.overrides
            or new_requires_python.is_empty()
            or requires_python.is_superset(new_requires_python)
        ):
            valid_deps.append(PythonRequirement.from_pyspec_set(requires_python))
        return valid_deps


@register_provider("reuse")
class ReusePinProvider(BaseProvider):
    """A provider that reuses preferred pins if possible.

    This is used to implement "add", "remove", and "reuse upgrade",
    where already-pinned candidates in lockfile should be preferred.
    """

    def __init__(
        self,
        repository: BaseRepository,
        allow_prereleases: bool | None = None,
        overrides: dict[str, str] | None = None,
        direct_minimal_versions: bool = False,
        *,
        locked_repository: LockedRepository | None = None,
        tracked_names: Iterable[str],
    ) -> None:
        super().__init__(
            repository=repository,
            allow_prereleases=allow_prereleases,
            overrides=overrides,
            direct_minimal_versions=direct_minimal_versions,
            locked_repository=locked_repository,
        )
        self.tracked_names = set(tracked_names)

    def iter_reuse_candidates(self, identifier: str, requirement: Requirement | None) -> Iterable[Candidate]:
        bare_name = strip_extras(identifier)[0]
        if bare_name in self.tracked_names or identifier not in self.locked_candidates:
            return []
        return sorted(self.locked_candidates[identifier], key=lambda c: c.version or "", reverse=True)

    def get_reuse_candidate(self, identifier: str, requirement: Requirement | None) -> Candidate | None:
        deprecation_warning(
            "The get_reuse_candidate method is deprecated, use iter_reuse_candidates instead.", stacklevel=2
        )
        return next(iter(self.iter_reuse_candidates(identifier, requirement)), None)

    def find_matches(
        self,
        identifier: str,
        requirements: Mapping[str, Iterator[Requirement]],
        incompatibilities: Mapping[str, Iterator[Candidate]],
    ) -> Callable[[], Iterator[Candidate]]:
        super_find = super().find_matches(identifier, requirements, incompatibilities)

        def matches_gen() -> Iterator[Candidate]:
            requested_req = next(filter(lambda r: r.is_named, requirements[identifier]), None)
            for pin in self.iter_reuse_candidates(identifier, requested_req):
                if identifier not in self.overrides and pin.req.is_named:
                    pin = pin.copy_with(min(requirements[identifier], key=self.requirement_preference))
                incompat = list(incompatibilities[identifier])
                pin._preferred = True  # type: ignore[attr-defined]
                if pin not in incompat and all(self.is_satisfied_by(r, pin) for r in requirements[identifier]):
                    yield pin
            yield from super_find()

        return matches_gen

    def _get_dependencies_from_repository(self, candidate: Candidate) -> tuple[list[Requirement], PySpecSet, str]:
        is_stable_metadata = candidate.req.is_named or (
            isinstance(candidate.req, VcsRequirement) and candidate.req.revision
        )
        if self.locked_repository is not None and is_stable_metadata:
            try:
                return self.locked_repository.get_dependencies(candidate)
            except CandidateNotFound:
                pass
        return super()._get_dependencies_from_repository(candidate)


@register_provider("eager")
class EagerUpdateProvider(ReusePinProvider):
    """A specialized provider to handle an "eager" upgrade strategy.

    An eager upgrade tries to upgrade not only packages specified, but also
    their dependencies (recursively). This contrasts to the "only-if-needed"
    default, which only promises to upgrade the specified package, and
    prevents touching anything else if at all possible.

    The provider is implemented as to keep track of all dependencies of the
    specified packages to upgrade, and free their pins when it has a chance.
    """

    def iter_reuse_candidates(self, identifier: str, requirement: Requirement | None) -> Iterable[Candidate]:
        if identifier in self.tracked_names:
            # If this is a tracked package, don't reuse its pinned version, so it can be upgraded.
            return []
        return super().iter_reuse_candidates(identifier, requirement)

    def get_dependencies(self, candidate: Candidate) -> list[Requirement]:
        # If this package is being tracked for upgrade, remove pins of its
        # dependencies, and start tracking these new packages.
        dependencies = super().get_dependencies(candidate)
        if self.identify(candidate) in self.tracked_names:
            for dependency in dependencies:
                if dependency.key:
                    self.tracked_names.add(dependency.key)
        return dependencies

    def get_preference(
        self,
        identifier: str,
        resolutions: Mapping[str, Candidate],
        candidates: Mapping[str, Iterator[Candidate]],
        information: Mapping[str, Iterator[RequirementInformation]],
        backtrack_causes: Sequence[RequirementInformation],
    ) -> tuple[Comparable, ...]:
        # Resolve tracking packages so we have a chance to unpin them first.
        (python, *others) = super().get_preference(identifier, resolutions, candidates, information, backtrack_causes)
        return (python, identifier not in self.tracked_names, *others)


@register_provider("reuse-installed")
class ReuseInstalledProvider(ReusePinProvider):
    """A provider that reuses installed packages if possible."""

    @cached_property
    def installed(self) -> WorkingSet:
        return self.repository.environment.get_working_set()

    def iter_reuse_candidates(self, identifier: str, requirement: Requirement | None) -> Iterable[Candidate]:
        key = strip_extras(identifier)[0]
        if key not in self.installed or requirement is None:
            return super().iter_reuse_candidates(identifier, requirement)
        else:
            dist = self.installed[key]
            return [Candidate(requirement, installed=dist)]



from __future__ import annotations

import dataclasses as dc
import sys
from functools import cached_property
from pathlib import Path

from pdm.models.in_process import get_sys_config_paths
from pdm.utils import find_python_in_path, get_venv_like_prefix

IS_WIN = sys.platform == "win32"
BIN_DIR = "Scripts" if IS_WIN else "bin"


def get_venv_python(venv: Path) -> Path:
    """Get the interpreter path inside the given venv."""
    suffix = ".exe" if IS_WIN else ""
    result = venv / BIN_DIR / f"python{suffix}"
    if IS_WIN and not result.exists():
        result = venv / "bin" / f"python{suffix}"  # for mingw64/msys2
        if result.exists():
            return result
        else:
            return venv / "python.exe"  # for conda
    return result


def is_conda_venv(root: Path) -> bool:
    return (root / "conda-meta").exists()


@dc.dataclass(frozen=True)
class VirtualEnv:
    root: Path
    is_conda: bool
    interpreter: Path

    @classmethod
    def get(cls, root: Path) -> VirtualEnv | None:
        path = get_venv_python(root)
        if not path.exists():
            return None
        return cls(root, is_conda_venv(root), path)

    @classmethod
    def from_interpreter(cls, interpreter: Path) -> VirtualEnv | None:
        root, is_conda = get_venv_like_prefix(interpreter)
        if root is not None:
            return cls(root, is_conda, interpreter)
        return None

    def env_vars(self) -> dict[str, str]:
        key = "CONDA_PREFIX" if self.is_conda else "VIRTUAL_ENV"
        return {key: str(self.root)}

    @cached_property
    def venv_config(self) -> dict[str, str]:
        venv_cfg = self.root / "pyvenv.cfg"
        if not venv_cfg.exists():
            return {}
        parsed: dict[str, str] = {}
        with venv_cfg.open(encoding="utf-8") as fp:
            for line in fp:
                if "=" in line:
                    k, v = line.split("=", 1)
                    k = k.strip().lower()
                    v = v.strip()
                    if k == "include-system-site-packages":
                        v = v.lower()
                    parsed[k] = v
        return parsed

    @property
    def include_system_site_packages(self) -> bool:
        return self.venv_config.get("include-system-site-packages") == "true"

    @cached_property
    def base_paths(self) -> list[str]:
        home = Path(self.venv_config["home"])
        base_executable = find_python_in_path(home) or find_python_in_path(home.parent)
        assert base_executable is not None
        paths = get_sys_config_paths(str(base_executable))
        return [paths["purelib"], paths["platlib"]]



from __future__ import annotations

import os
import shutil
from functools import cached_property
from pathlib import Path
from typing import Any, ClassVar, ContextManager

from pdm.termui import logger


class CachedPackage:
    """A package cached in the central package store.
    The directory name is similar to wheel's filename:

        $PACKAGE_ROOT/<checksum[:2]>/<dist_name>-<version>-<impl>-<abi>-<plat>/

    The checksum is stored in a file named `.checksum` under the directory.

    Under the directory there could be a text file named `.referrers`.
    Each line of the file is a distribution path that refers to this package.
    *Only wheel installations will be cached*
    """

    cache_files: ClassVar[tuple[str, ...]] = (".lock", ".checksum", ".referrers")
    """List of files storing cache metadata and not being part of the package"""

    def __init__(self, path: str | Path, original_wheel: Path | None = None) -> None:
        self.path = Path(os.path.normcase(os.path.expanduser(path))).resolve()
        self.original_wheel = original_wheel
        self._referrers: set[str] | None = None

    def lock(self) -> ContextManager[Any]:
        import filelock

        return filelock.FileLock(self.path / ".lock")

    @cached_property
    def checksum(self) -> str:
        """The checksum of the path"""
        return self.path.joinpath(".checksum").read_text().strip()

    @cached_property
    def dist_info(self) -> Path:
        """The dist-info directory of the wheel"""
        from installer.exceptions import InvalidWheelSource

        try:
            return next(self.path.glob("*.dist-info"))
        except StopIteration:
            raise InvalidWheelSource(f"The wheel doesn't contain metadata {self.path!r}") from None

    @property
    def referrers(self) -> set[str]:
        """A set of entries in referrers file"""
        if self._referrers is None:
            filepath = self.path / ".referrers"
            if not filepath.is_file():
                return set()
            self._referrers = {
                line.strip()
                for line in filepath.read_text("utf8").splitlines()
                if line.strip() and os.path.exists(line.strip())
            }
        return self._referrers

    def add_referrer(self, path: str) -> None:
        """Add a new referrer"""
        path = os.path.normcase(os.path.expanduser(os.path.abspath(path)))
        referrers = self.referrers | {path}
        (self.path / ".referrers").write_text("\n".join(sorted(referrers)) + "\n", "utf8")
        self._referrers = None

    def remove_referrer(self, path: str) -> None:
        """Remove a referrer"""
        path = os.path.normcase(os.path.expanduser(os.path.abspath(path)))
        referrers = self.referrers - {path}
        (self.path / ".referrers").write_text("\n".join(referrers) + "\n", "utf8")
        self._referrers = None

    def cleanup(self) -> None:
        logger.info("Clean up cached package %s", self.path)
        shutil.rmtree(self.path)



"""
The signal definition for PDM.

Example:
    ```python
    from pdm.signals import post_init, post_install

    def on_post_init(project):
        project.core.ui.echo("Project initialized")
    # Connect to the signal
    post_init.connect(on_post_init)
    # Or use as a decorator
    @post_install.connect
    def on_post_install(project, candidates, dry_run):
        project.core.ui.echo("Project install succeeded")
    ```
"""

from blinker import NamedSignal, Namespace

pdm_signals = Namespace()

post_init: NamedSignal = pdm_signals.signal("post_init")
"""Called after a project is initialized.
Args:
    project (Project): The project object
"""
pre_lock: NamedSignal = pdm_signals.signal("pre_lock")
"""Called before a project is locked.
Args:
    project (Project): The project object
    requirements (list[Requirement]): The requirements to lock
    dry_run (bool): If true, won't perform any actions
"""
post_lock: NamedSignal = pdm_signals.signal("post_lock")
"""Called after a project is locked.

Args:
    project (Project): The project object
    resolution (dict[str, list[Candidate]]): The resolved candidates
    dry_run (bool): If true, won't perform any actions
"""
pre_install: NamedSignal = pdm_signals.signal("pre_install")
"""Called before a project is installed.

Args:
    project (Project): The project object
    packages (list[Package]): The packages to install
    dry_run (bool): If true, won't perform any actions
"""
post_install: NamedSignal = pdm_signals.signal("post_install")
"""Called after a project is installed.

Args:
    project (Project): The project object
    packages (list[Package]): The packages installed
    dry_run (bool): If true, won't perform any actions
"""
pre_build: NamedSignal = pdm_signals.signal("pre_build")
"""Called before a project is built.

Args:
    project (Project): The project object
    dest (str): The destination location
    config_settings (dict[str, str]|None): Additional config settings passed via args
"""
post_build: NamedSignal = pdm_signals.signal("post_build")
"""Called after a project is built.

Args:
    project (Project): The project object
    artifacts (Sequence[str]): The locations of built artifacts
    config_settings (dict[str, str]|None): Additional config settings passed via args
"""
pre_publish: NamedSignal = pdm_signals.signal("pre_publish")
"""Called before a project is published.

Args:
    project (Project): The project object
"""
post_publish: NamedSignal = pdm_signals.signal("post_publish")
"""Called after a project is published.

Args:
    project (Project): The project object
"""
pre_run: NamedSignal = pdm_signals.signal("pre_run")
"""Called before any run.

Args:
    project (Project): The project object
    script (str): the script name
    args (Sequence[str]): the command line provided arguments
"""
post_run: NamedSignal = pdm_signals.signal("post_run")
"""Called after any run.

Args:
    project (Project): The project object
    script (str): the script name
    args (Sequence[str]): the command line provided arguments
"""
pre_script: NamedSignal = pdm_signals.signal("pre_script")
"""Called before any script.

Args:
    project (Project): The project object
    script (str): the script name
    args (Sequence[str]): the command line provided arguments
"""
post_script: NamedSignal = pdm_signals.signal("post_script")
"""Called after any script.

Args:
    project (Project): The project object
    script (str): the script name
    args (Sequence[str]): the command line provided arguments
"""
post_use: NamedSignal = pdm_signals.signal("post_use")
"""Called after use switched to a new Python version.

Args:
    project (Project): The project object
    python (PythonInfo): Information about the new Python interpreter
"""
pre_invoke: NamedSignal = pdm_signals.signal("pre_invoke")
"""Called before any command is invoked.

Args:
    project (Project): The project object
    command (str | None): the command name
    options (Namespace): the parsed arguments
"""



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Iterable

import tomlkit
from dep_logic.markers import AnyMarker, BaseMarker, MarkerUnion, parse_marker

from pdm.exceptions import ProjectError
from pdm.formats.base import make_array, make_inline_table
from pdm.models.candidates import Candidate
from pdm.models.requirements import FileRequirement, VcsRequirement
from pdm.project.lockfile import FLAG_INHERIT_METADATA
from pdm.utils import cd, normalize_name

if TYPE_CHECKING:
    from pdm.models.markers import Marker
    from pdm.models.repositories.lock import LockedRepository, Package
    from pdm.project import Project


def _group_sort_key(group: str) -> tuple[bool, str]:
    return group != "default", group


class PyLockConverter:
    lock_version = "1.0"

    def __init__(self, project: Project, locked_repository: LockedRepository) -> None:
        self.project = project
        self.locked_repository = locked_repository

    def make_package(self, package: Package) -> dict[str, Any]:
        req = package.candidate.req
        candidate = package.candidate
        result: dict[str, Any] = {
            "name": candidate.req.key,
            "version": candidate.version,
        }
        if candidate.requires_python:
            result["requires-python"] = candidate.requires_python
        if isinstance(req, VcsRequirement):
            result["vcs"] = {
                "type": req.vcs,
                "url": req.repo,
                "commit-id": candidate.get_revision(),
            }
            if req.ref:
                result["vcs"]["requested-revision"] = req.ref
            if req.subdirectory:
                result["vcs"]["subdirectory"] = req.subdirectory
        elif isinstance(req, FileRequirement):
            if req.is_local_dir:
                result["directory"] = {
                    "path": req.str_path,
                    "editable": req.editable,
                }
                if req.subdirectory:
                    result["directory"]["subdirectory"] = req.subdirectory
            else:
                archive: dict[str, Any]
                archive = result["archive"] = {"url": req.get_full_url()}
                if req.path:
                    archive["path"] = req.str_path
                for hash_item in candidate.hashes:
                    hash_type, hash_value = hash_item["hash"].split(":", 1)
                    archive.setdefault("hashes", tomlkit.inline_table())[hash_type] = hash_value
        else:
            wheels: list[dict[str, Any]] = tomlkit.array().multiline(True)
            for hash_item in candidate.hashes:
                hash_type, hash_value = hash_item["hash"].split(":", 1)
                if hash_item.get("file", "").endswith(".whl") or hash_item.get("url", "").endswith(".whl"):
                    wheel: dict[str, Any] = {}
                    if "file" in hash_item:
                        wheel["name"] = hash_item["file"]
                    if "url" in hash_item:
                        wheel["url"] = hash_item["url"]
                    wheel["hashes"] = make_inline_table({hash_type: hash_value})
                    wheels.append(wheel)
                else:
                    sdist: dict[str, Any] = {}
                    if "file" in hash_item:
                        sdist["name"] = hash_item["file"]
                    if "url" in hash_item:
                        sdist["url"] = hash_item["url"]
                    sdist["hashes"] = make_inline_table({hash_type: hash_value})
                    result["sdist"] = make_inline_table(sdist)
            if wheels:
                result["wheels"] = wheels

            if package.dependencies is not None:
                result["tool"] = {"pdm": {"dependencies": make_array(package.dependencies, multiline=True)}}

        return result

    def _populate_hashes(self, packages: Iterable[Package]) -> None:
        candidates: list[Candidate] = []
        for package in packages:
            if not package.candidate.req.is_named or package.candidate.req.extras:
                continue
            hashes = package.candidate.hashes
            if all("url" in hash_item for hash_item in hashes):
                continue
            package.candidate.hashes.clear()
            candidates.append(package.candidate)
        if candidates:
            with self.project.core.ui.open_spinner("Fetching package file URLs"):
                repo = self.project.get_repository()
                repo.fetch_hashes(candidates)

    def convert(self, all_groups: Iterable[str] | None = None) -> dict[str, Any]:
        doc = tomlkit.document()
        project = self.project
        lockfile = project.lockfile
        if FLAG_INHERIT_METADATA not in lockfile.strategy:
            raise ProjectError("inherit_metadata strategy is required for pylock format")
        repository = self.locked_repository
        if all_groups is None:
            all_groups = lockfile.groups
        if all_groups is None:
            all_groups = list(project.iter_groups())
        extras, groups = self.project.split_extras_groups(list(all_groups))
        env_markers: list[Marker] = []
        for target in repository.targets:
            env_markers.append(target.markers_with_python())

        doc.update(
            {
                "lock-version": self.lock_version,
                "requires-python": str(project.python_requires),
                "environments": make_array([str(marker) for marker in env_markers], multiline=True),
                "extras": sorted(extras),
                "dependency-groups": sorted(groups, key=_group_sort_key),
                "default-groups": ["default"],
                "created-by": "pdm",
            }
        )
        packages = doc.setdefault("packages", tomlkit.aot())

        with cd(project.root):
            self._populate_hashes(repository.packages.values())
            for package in repository.packages.values():
                if package.candidate.req.extras:
                    continue
                package_table = self.make_package(package)
                selection_markers: list[BaseMarker] = []
                if not package.marker.is_any():
                    selection_markers.append(package.marker)
                for item in sorted(package.candidate.req.groups, key=_group_sort_key):
                    item = normalize_name(item)
                    if item in extras:
                        selection_markers.append(parse_marker(f"'{item}' in extras"))
                    else:
                        selection_markers.append(parse_marker(f"'{item}' in dependency_groups"))
                marker = MarkerUnion.of(*selection_markers) if selection_markers else AnyMarker()
                if package.candidate.req.marker is not None:
                    marker = package.candidate.req.marker.inner & marker
                if not marker.is_any():
                    package_table["marker"] = str(marker)
                packages.append(package_table)

        doc["tool"] = {
            "pdm": {
                "hashes": make_inline_table({"sha256": project.pyproject.content_hash()}),
                "targets": [spec.as_dict() for spec in repository.targets],
            },
        }

        return doc



from __future__ import annotations

import contextlib
import hashlib
import itertools
import operator
import os
import shutil
import sys
from copy import deepcopy
from functools import cached_property, reduce
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable, Iterable, Mapping, Sequence, cast

import tomlkit
from pbs_installer import PythonVersion

from pdm._types import NotSet, NotSetType, RepositoryConfig
from pdm.compat import CompatibleSequence, tomllib
from pdm.exceptions import NoPythonVersion, PdmUsageError, ProjectError
from pdm.models.backends import DEFAULT_BACKEND, BuildBackend, get_backend_by_spec
from pdm.models.caches import PackageCache
from pdm.models.markers import EnvSpec
from pdm.models.python import PythonInfo
from pdm.models.repositories import BaseRepository, LockedRepository
from pdm.models.requirements import Requirement, parse_line, parse_requirement, strip_extras
from pdm.models.specifiers import PySpecSet
from pdm.project.config import Config, ensure_boolean
from pdm.project.lockfile import FLAG_INHERIT_METADATA, Lockfile, load_lockfile
from pdm.project.project_file import PyProject
from pdm.utils import (
    cd,
    deprecation_warning,
    expand_env_vars_in_auth,
    find_project_root,
    find_python_in_path,
    get_all_installable_python_versions,
    get_class_init_params,
    is_conda_base,
    is_conda_base_python,
    is_path_relative_to,
    normalize_name,
)

if TYPE_CHECKING:
    from findpython import Finder

    from pdm.core import Core
    from pdm.environments import BaseEnvironment
    from pdm.installers.base import BaseSynchronizer
    from pdm.models.caches import CandidateInfoCache, HashCache, WheelCache
    from pdm.models.candidates import Candidate
    from pdm.resolver.base import Resolver
    from pdm.resolver.providers import BaseProvider
    from pdm.resolver.reporters import RichLockReporter


PYENV_ROOT = os.path.expanduser(os.getenv("PYENV_ROOT", "~/.pyenv"))


class Project:
    """Core project class.

    Args:
        core: The core instance.
        root_path: The root path of the project.
        is_global: Whether the project is global.
        global_config: The path to the global config file.
    """

    PYPROJECT_FILENAME = "pyproject.toml"

    def __init__(
        self,
        core: Core,
        root_path: str | Path | None,
        is_global: bool = False,
        global_config: str | Path | None = None,
    ) -> None:
        import platformdirs

        self._lockfile: Lockfile | None = None
        self._environment: BaseEnvironment | None = None
        self._python: PythonInfo | None = None
        self._cache_dir: Path | None = None
        self.core = core

        if global_config is None:
            global_config = platformdirs.user_config_path("pdm") / "config.toml"
        self.global_config = Config(Path(global_config), is_global=True)
        global_project = Path(self.global_config["global_project.path"]).expanduser()

        if root_path is None:
            root_path = find_project_root() if not is_global else global_project
        if (
            not is_global
            and root_path is None
            and self.global_config["global_project.fallback"]
            and not is_conda_base()
        ):
            root_path = global_project
            is_global = True
            if self.global_config["global_project.fallback_verbose"]:
                self.core.ui.info("Project is not found, fallback to the global project")

        self.root: Path = Path(root_path or "").absolute()
        self.is_global = is_global
        self.enable_write_lockfile = os.getenv("PDM_FROZEN_LOCKFILE", os.getenv("PDM_NO_LOCK", "0")).lower() not in (
            "1",
            "true",
        )
        self.init_global_project()

    def __repr__(self) -> str:
        return f"<Project '{self.root.as_posix()}'>"

    @cached_property
    def cache_dir(self) -> Path:
        return Path(self.config.get("cache_dir", "")).expanduser()

    @cached_property
    def pyproject(self) -> PyProject:
        return PyProject(self.root / self.PYPROJECT_FILENAME, ui=self.core.ui)

    @property
    def lockfile(self) -> Lockfile:
        if self._lockfile is None:
            enable_pylock = self.config["lock.format"] == "pylock"
            if (path := self.root / "pylock.toml").exists() and enable_pylock:
                self.set_lockfile(path)
            elif (path := self.root / "pdm.lock").exists():
                if enable_pylock:  # pragma: no cover
                    self.core.ui.warn(
                        "`lock.format` is set to pylock but pylock.toml is not found, using pdm.lock instead. "
                        "You can generate pylock with `pdm export -f pylock -o pylock.toml`."
                    )
                self.set_lockfile(path)
            else:
                file_path = "pylock.toml" if enable_pylock else "pdm.lock"
                self.set_lockfile(self.root / file_path)
            assert self._lockfile is not None
        return self._lockfile

    def set_lockfile(self, path: str | Path) -> None:
        self._lockfile = load_lockfile(self, path)
        if self.config.get("use_uv"):
            self._lockfile.default_strategies.discard(FLAG_INHERIT_METADATA)
        if not self.config["strategy.inherit_metadata"]:
            self._lockfile.default_strategies.discard(FLAG_INHERIT_METADATA)

    @cached_property
    def config(self) -> Mapping[str, Any]:
        """A read-only dict configuration"""
        import collections

        return collections.ChainMap(self.project_config, self.global_config)

    @property
    def scripts(self) -> dict[str, str | dict[str, str]]:
        return self.pyproject.settings.get("scripts", {})

    @cached_property
    def project_config(self) -> Config:
        """Read-and-writable configuration dict for project settings"""
        config = Config(self.root / "pdm.toml")
        # TODO: for backward compatibility, remove this in the future
        if self.root.joinpath(".pdm.toml").exists():
            legacy_config = Config(self.root / ".pdm.toml").self_data
            config.update((k, v) for k, v in legacy_config.items() if k != "python.path")
        return config

    @property
    def name(self) -> str:
        return cast(str, self.pyproject.metadata.get("name"))

    @property
    def python(self) -> PythonInfo:
        if not self._python:
            python = self.resolve_interpreter()
            if python.major < 3:
                raise PdmUsageError(
                    "Python 2.7 has reached EOL and PDM no longer supports it. "
                    "Please upgrade your Python to 3.6 or later.",
                )
            if self.is_global and is_conda_base_python(python.path):  # pragma: no cover
                raise PdmUsageError("Can't use global project in conda base environment since it is managed by conda")
            self._python = python
        return self._python

    @python.setter
    def python(self, value: PythonInfo) -> None:
        self._python = value
        self._saved_python = value.path.as_posix()

    @property
    def _saved_python(self) -> str | None:
        if os.getenv("PDM_PYTHON"):
            return os.getenv("PDM_PYTHON")
        with contextlib.suppress(FileNotFoundError):
            return self.root.joinpath(".pdm-python").read_text("utf-8").strip()
        with contextlib.suppress(FileNotFoundError):
            # TODO: remove this in the future
            with self.root.joinpath(".pdm.toml").open("rb") as fp:
                data = tomllib.load(fp)
                if data.get("python", {}).get("path"):
                    return data["python"]["path"]
        return None

    @_saved_python.setter
    def _saved_python(self, value: str | None) -> None:
        self.root.mkdir(parents=True, exist_ok=True)
        python_file = self.root.joinpath(".pdm-python")
        if value is None:
            with contextlib.suppress(FileNotFoundError):
                python_file.unlink()
            return
        python_file.write_text(value, "utf-8")

    def resolve_interpreter(self) -> PythonInfo:
        """Get the Python interpreter path."""
        from pdm.cli.commands.venv.utils import iter_venvs
        from pdm.models.venv import get_venv_python

        def match_version(python: PythonInfo) -> bool:
            return python.valid and self.python_requires.contains(python.version, True)

        def note(message: str) -> None:
            if not self.is_global:
                self.core.ui.info(message)

        def is_active_venv(python: PythonInfo) -> bool:
            if not (venv := os.getenv("VIRTUAL_ENV", os.getenv("CONDA_PREFIX"))):
                return False
            return is_path_relative_to(python.executable, venv)

        config = self.config
        saved_path = self._saved_python
        if saved_path and not ensure_boolean(os.getenv("PDM_IGNORE_SAVED_PYTHON")):
            python = PythonInfo.from_path(saved_path)
            if match_version(python):
                return python
            elif not python.valid:
                note("The saved Python interpreter does not exist or broken. Trying to find another one.")
            else:
                note(
                    "The saved Python interpreter doesn't match the project's requirement. Trying to find another one."
                )
            self._saved_python = None  # Clear the saved path if it doesn't match

        if config.get("python.use_venv") and not self.is_global:
            # Resolve virtual environments from env-vars
            ignore_active_venv = ensure_boolean(os.getenv("PDM_IGNORE_ACTIVE_VENV"))
            venv_in_env = os.getenv("VIRTUAL_ENV", os.getenv("CONDA_PREFIX"))
            # We don't auto reuse conda's base env since it may cause breakage when removing packages.
            if not ignore_active_venv and venv_in_env and not is_conda_base():
                python = PythonInfo.from_path(get_venv_python(Path(venv_in_env)))
                if match_version(python):
                    note(
                        f"Inside an active virtualenv [success]{venv_in_env}[/], reusing it.\n"
                        "Set env var [success]PDM_IGNORE_ACTIVE_VENV[/] to ignore it."
                    )
                    return python
            # otherwise, get a venv associated with the project
            for _, venv in iter_venvs(self):
                python = PythonInfo.from_path(venv.interpreter)
                if match_version(python) and not (ignore_active_venv and is_active_venv(python)):
                    note(f"Virtualenv [success]{venv.root}[/] is reused.")
                    self.python = python
                    return python

            if not self.root.joinpath("__pypackages__").exists():
                self.core.ui.warn(
                    f"Project requires a python version of {self.python_requires}, "
                    f"The virtualenv is being created for you as it cannot be matched to the right version."
                )
                note("python.use_venv is on, creating a virtualenv for this project...")
                venv_path = self._create_virtualenv()
                self.python = PythonInfo.from_path(get_venv_python(venv_path))
                return self.python

        if self.root.joinpath("__pypackages__").exists() or not config["python.use_venv"] or self.is_global:
            for py_version in self.iter_interpreters(
                filter_func=match_version, respect_version_file=config["python.use_python_version"]
            ):
                note("[success]__pypackages__[/] is detected, using the PEP 582 mode")
                self.python = py_version
                return py_version

        raise NoPythonVersion(f"No Python that satisfies {self.python_requires} is found on the system.")

    def get_environment(self) -> BaseEnvironment:
        from pdm.environments import PythonEnvironment, PythonLocalEnvironment

        """Get the environment selected by this project"""

        if self.is_global:
            env = PythonEnvironment(self)
            # Rewrite global project's python requires to be
            # compatible with the exact version
            env.python_requires = PySpecSet(f"=={self.python.version}")
            return env

        return (
            PythonEnvironment(self)
            if self.config["python.use_venv"] and self.python.get_venv() is not None
            else PythonLocalEnvironment(self)
        )

    def _create_virtualenv(self, python: str | None = None) -> Path:
        from pdm.cli.commands.venv.backends import BACKENDS

        backend: str = self.config["venv.backend"]
        if backend == "virtualenv" and self.config["use_uv"]:
            backend = "uv"
        venv_backend = BACKENDS[backend](self, python)
        path = venv_backend.create(
            force=True,
            in_project=self.config["venv.in_project"],
            prompt=self.config["venv.prompt"],
            with_pip=self.config["venv.with_pip"],
        )
        self.core.ui.echo(f"Virtualenv is created successfully at [success]{path}[/]", err=True)
        return path

    @property
    def environment(self) -> BaseEnvironment:
        if not self._environment:
            self._environment = self.get_environment()
        return self._environment

    @environment.setter
    def environment(self, value: BaseEnvironment | None) -> None:
        self._environment = value

    @property
    def python_requires(self) -> PySpecSet:
        return PySpecSet(self.pyproject.metadata.get("requires-python", ""))

    def get_dependencies(
        self, group: str | None = None, all_dependencies: dict[str, list[Requirement]] | None = None
    ) -> Sequence[Requirement]:
        group = normalize_name(group or "default")
        if all_dependencies is None:
            all_dependencies = self._resolve_dependencies([group])
        if group not in all_dependencies:
            raise ProjectError(f"Dependency group {group} does not exist")
        return CompatibleSequence(all_dependencies[group])

    def iter_groups(self) -> Iterable[str]:
        groups = {"default"}
        if self.pyproject.metadata.get("optional-dependencies"):
            groups.update(self.pyproject.metadata["optional-dependencies"].keys())
        groups.update(self.pyproject._data.get("dependency-groups", {}).keys())
        groups.update(self.pyproject.settings.get("dev-dependencies", {}).keys())
        return {normalize_name(g) for g in groups}

    def _resolve_dependencies(
        self, requested_groups: list[str] | None = None, include_referred: bool = True
    ) -> dict[str, list[Requirement]]:
        """Resolve dependencies for the given groups, and return a list of requirements for each group.

        The .groups attribute will be set to all that refers this requirement directly or indirectly.
        If `include_referred` is True, all self-references and `include-group` will be expanded to
        corresponding requirements. Otherwise, each group only contains explicitly defined requirements.
        """

        def _get_dependencies(group: str) -> tuple[list[Requirement], set[str]]:
            in_metadata = group in metadata_dependencies
            collected_deps: list[str] = []
            referred: set[str] = set()
            deps = metadata_dependencies.get(group, []) if in_metadata else dev_dependencies[group]
            for item in deps:
                if isinstance(item, str):
                    try:
                        name, extras = strip_extras(item)
                    except AssertionError:
                        pass
                    else:
                        if normalize_name(name) == project_name:
                            if extras:
                                allowed = (
                                    set(metadata_dependencies)
                                    if in_metadata
                                    else {*metadata_dependencies, *dev_dependencies}
                                )
                                extras = tuple(normalize_name(extra) for extra in extras)
                                not_allowed = set(extras) - allowed
                                if not_allowed:
                                    raise ProjectError(
                                        f"Optional dependency group '{group}' cannot "
                                        f"include non-existing extras: [{','.join(not_allowed)}]"
                                    )
                                referred.update(extras)
                            continue
                    collected_deps.append(item)
                elif not in_metadata and isinstance(item, dict):
                    if tuple(item.keys()) != ("include-group",):
                        raise ProjectError(f"Invalid dependency group item: {item}")
                    include_group = normalize_name(item["include-group"])
                    if include_group not in dev_dependencies:
                        raise ProjectError(f"Missing group '{include_group}' in `include-group`")
                    referred.add(include_group)
                else:
                    raise ProjectError(f"Invalid dependency in group {group}: {item}")
            result: list[Requirement] = []
            with cd(self.root):
                for line in collected_deps:
                    if line.startswith("-e ") and in_metadata:
                        self.core.ui.warn(
                            f"Skipping editable dependency [b]{line}[/] in the"
                            r" [success]\[project][/] table. Please move it to the "
                            r"[success]\[tool.pdm.dev-dependencies][/] table"
                        )
                        continue
                    req = parse_line(line)
                    req.groups = [group]
                    # make editable packages behind normal ones to override correctly.
                    result.append(req)
            return result, referred

        if requested_groups is None:
            requested_groups = list(self.iter_groups())
        requested_groups = [normalize_name(g) for g in requested_groups]
        referred_groups: dict[str, set[str]] = {}
        metadata_dependencies = {
            normalize_name(k): v for k, v in self.pyproject.metadata.get("optional-dependencies", {}).items()
        }
        if "default" in metadata_dependencies:  # pragma: no cover
            raise ProjectError(
                "'default' is reserved by the main dependencies and is not allowed in optional dependencies."
            )
        metadata_dependencies["default"] = self.pyproject.metadata.get("dependencies", [])
        dev_dependencies = self.pyproject.dev_dependencies
        if "default" in dev_dependencies:  # pragma: no cover
            raise ProjectError(
                "'default' is reserved by the main dependencies and is not allowed in dependency groups."
            )
        group_deps: dict[str, list[Requirement]] = {}
        project_name = normalize_name(self.name) if self.name else None
        for group in requested_groups:
            deps, referred = _get_dependencies(group)
            group_deps[group] = deps
            if referred:
                referred_groups[group] = referred
        extra_deps: dict[str, list[Requirement]] = {}
        while referred_groups:
            updated = False
            ref_iter = list(referred_groups.items())
            for group, referred in ref_iter:
                for ref in list(referred):
                    if ref not in requested_groups:
                        deps, r = _get_dependencies(ref)
                        group_deps[ref] = deps
                        if r:
                            referred_groups[ref] = r
                            # append to the ref_iter to process later
                            ref_iter.append((ref, r))
                        requested_groups.append(ref)
                    if ref in referred_groups:  # not resolved yet
                        continue
                    extra_deps.setdefault(group, []).extend(group_deps[ref])
                    for req in itertools.chain(group_deps[ref], extra_deps.get(ref, [])):
                        if group not in req.groups:
                            req.groups.append(group)
                    referred.remove(ref)
                    updated = True
                if not referred:
                    referred_groups.pop(group)
            if not updated:
                raise ProjectError(f"Cyclic dependency group include detected: {set(referred_groups)}")
        if include_referred:
            for group, deps in extra_deps.items():
                group_deps[group].extend(deps)
        return group_deps

    @property
    def all_dependencies(self) -> dict[str, Sequence[Requirement]]:
        return {k: CompatibleSequence(v) for k, v in self._resolve_dependencies(include_referred=False).items()}

    @property
    def default_source(self) -> RepositoryConfig:
        """Get the default source from the pypi setting"""
        config = RepositoryConfig(
            config_prefix="pypi",
            name="pypi",
            url=self.config["pypi.url"],
            verify_ssl=self.config["pypi.verify_ssl"],
            username=self.config.get("pypi.username"),
            password=self.config.get("pypi.password"),
            ca_certs=self.config.get("pypi.ca_certs"),
            client_cert=self.config.get("pypi.client_cert"),
            client_key=self.config.get("pypi.client_key"),
        )
        return config

    @property
    def sources(self) -> list[RepositoryConfig]:
        return self.get_sources(include_stored=not self.config.get("pypi.ignore_stored_index", False))

    def get_sources(self, expand_env: bool = True, include_stored: bool = False) -> list[RepositoryConfig]:
        result: dict[str, RepositoryConfig] = {}
        for source in self.pyproject.settings.get("source", []):
            result[source["name"]] = RepositoryConfig(**source, config_prefix="pypi")

        def merge_sources(other_sources: Iterable[RepositoryConfig]) -> None:
            for source in other_sources:
                name = source.name
                if name in result:
                    result[name].passive_update(source)
                elif include_stored:
                    result[name] = source

        merge_sources(self.project_config.iter_sources())
        merge_sources(self.global_config.iter_sources())
        if "pypi" in result:
            result["pypi"].passive_update(self.default_source)
        elif include_stored:
            # put pypi source at the beginning
            result = {"pypi": self.default_source, **result}

        sources: list[RepositoryConfig] = []
        for source in result.values():
            if not source.url:
                continue
            if expand_env:
                source.url = DEFAULT_BACKEND(self.root).expand_line(expand_env_vars_in_auth(source.url))
            sources.append(source)
        return sources

    def get_repository(
        self,
        cls: type[BaseRepository] | None = None,
        ignore_compatibility: bool | NotSetType = NotSet,
        env_spec: EnvSpec | None = None,
    ) -> BaseRepository:
        """Get the repository object"""
        if cls is None:
            cls = self.core.repository_class
        sources = self.sources or []
        params = get_class_init_params(cls)
        if "env_spec" in params:
            return cls(sources, self.environment, env_spec=env_spec)
        else:
            return cls(sources, self.environment, ignore_compatibility=ignore_compatibility)

    def get_locked_repository(self, env_spec: EnvSpec | None = None) -> LockedRepository:
        try:
            lockfile = self.lockfile.open_for_read()
        except ProjectError:
            lockfile = {}

        return LockedRepository(lockfile, self.sources, self.environment, env_spec=env_spec)

    def split_extras_groups(self, all_groups: list[str]) -> tuple[list[str], list[str]]:
        """Split the groups into extras and non-extras."""
        extras: list[str] = []
        groups: list[str] = []
        optional_groups = {normalize_name(group) for group in self.pyproject.metadata.get("optional-dependencies", [])}
        for group in all_groups:
            if group in optional_groups:
                extras.append(group)
            else:
                groups.append(group)
        return extras, groups

    @property
    def locked_repository(self) -> LockedRepository:
        deprecation_warning("Project.locked_repository is deprecated, use Project.get_locked_repository() instead", 2)
        return self.get_locked_repository()

    def get_provider(
        self,
        strategy: str = "all",
        tracked_names: Iterable[str] | None = None,
        for_install: bool = False,
        ignore_compatibility: bool | NotSetType = NotSet,
        direct_minimal_versions: bool = False,
        env_spec: EnvSpec | None = None,
        locked_repository: LockedRepository | None = None,
    ) -> BaseProvider:
        """Build a provider class for resolver.

        :param strategy: the resolve strategy
        :param tracked_names: the names of packages that needs to update
        :param for_install: if the provider is for install
        :param ignore_compatibility: if the provider should ignore the compatibility when evaluating candidates
        :param direct_minimal_versions: if the provider should prefer minimal versions instead of latest
        :returns: The provider object
        """

        import inspect

        from pdm.resolver.providers import get_provider

        if env_spec is None:
            env_spec = (
                self.environment.allow_all_spec if ignore_compatibility in (True, NotSet) else self.environment.spec
            )
        repo_params = inspect.signature(self.get_repository).parameters
        if "env_spec" in repo_params:
            repository = self.get_repository(env_spec=env_spec)
        else:  # pragma: no cover
            repository = self.get_repository(ignore_compatibility=ignore_compatibility)
        if locked_repository is None:
            try:
                locked_repository = self.get_locked_repository(env_spec)
            except Exception:  # pragma: no cover
                if strategy != "all":
                    self.core.ui.warn("Unable to reuse the lock file as it is not compatible with PDM")

        provider_class = get_provider(strategy)
        params: dict[str, Any] = {}
        if strategy != "all":
            params["tracked_names"] = [strip_extras(name)[0] for name in tracked_names or ()]
        if "locked_repository" in inspect.signature(provider_class).parameters:
            params["locked_repository"] = locked_repository
        else:
            locked_candidates: dict[str, list[Candidate]] = (
                {} if locked_repository is None else locked_repository.all_candidates
            )
            params["locked_candidates"] = locked_candidates
        return provider_class(repository=repository, direct_minimal_versions=direct_minimal_versions, **params)

    def get_reporter(
        self, requirements: list[Requirement], tracked_names: Iterable[str] | None = None
    ) -> RichLockReporter:  # pragma: no cover
        """Return the reporter object to construct a resolver.

        :param requirements: requirements to resolve
        :param tracked_names: the names of packages that needs to update
        :param spinner: optional spinner object
        :returns: a reporter
        """
        from pdm.resolver.reporters import RichLockReporter

        return RichLockReporter(requirements, self.core.ui)

    def write_lockfile(
        self, toml_data: Any = None, show_message: bool = True, write: bool = True, **_kwds: Any
    ) -> None:
        """Write the lock file to disk."""
        if _kwds:  # pragma: no cover
            deprecation_warning("Extra arguments have been moved to `format_lockfile` function", stacklevel=2)
        if toml_data is not None:  # pragma: no cover
            deprecation_warning(
                "Passing toml_data to write_lockfile is deprecated, please use `format_lockfile` instead", stacklevel=2
            )
            self.lockfile.set_data(toml_data)
        self.lockfile.update_hash(self.pyproject.content_hash("sha256"))
        if write and self.enable_write_lockfile:
            self.lockfile.write(show_message)

    def make_self_candidate(self, editable: bool = True) -> Candidate:
        from unearth import Link

        from pdm.models.candidates import Candidate

        req = parse_requirement(self.root.as_uri(), editable)
        assert self.name
        req.name = self.name
        can = Candidate(req, name=self.name, link=Link.from_path(self.root))
        can.prepare(self.environment).metadata
        return can

    def is_lockfile_hash_match(self) -> bool:
        algo, hash_value = self.lockfile.hash
        if not hash_value:
            return False
        content_hash = self.pyproject.content_hash(algo)
        return content_hash == hash_value

    def use_pyproject_dependencies(
        self, group: str, dev: bool = False
    ) -> tuple[list[str], Callable[[list[str]], None]]:
        """Get the dependencies array and setter in the pyproject.toml
        Return a tuple of two elements, the first is the dependencies array,
        and the second value is a callable to set the dependencies array back.
        """
        from pdm.formats.base import make_array

        def update_dev_dependencies(deps: list[str]) -> None:
            from tomlkit.container import OutOfOrderTableProxy

            dependency_groups: list[str | dict[str, str]] = tomlkit.array().multiline(True)
            dev_dependencies: list[str] = tomlkit.array().multiline(True)
            for dep in deps:
                if isinstance(dep, str) and dep.startswith("-e"):
                    dev_dependencies.append(dep)
                else:
                    dependency_groups.append(dep)
            if dependency_groups:
                self.pyproject.dependency_groups[group] = dependency_groups
            else:
                self.pyproject.dependency_groups.pop(group, None)
            if dev_dependencies:
                settings.setdefault("dev-dependencies", {})[group] = dev_dependencies
            else:
                settings.setdefault("dev-dependencies", {}).pop(group, None)
            if isinstance(self.pyproject._data["tool"], OutOfOrderTableProxy):
                # In case of a separate table, we have to remove and re-add it to make the write correct.
                # This may change the order of tables in the TOML file, but it's the best we can do.
                # see bug pdm-project/pdm#2056 for details
                del self.pyproject._data["tool"]["pdm"]
                self.pyproject._data["tool"]["pdm"] = settings

        metadata, settings = self.pyproject.metadata, self.pyproject.settings
        if group == "default":
            return metadata.get("dependencies", tomlkit.array()), lambda x: metadata.__setitem__("dependencies", x)
        dev_dependencies = deepcopy(self.pyproject._data.get("dependency-groups", {}))
        for dev_group, items in self.pyproject.settings.get("dev-dependencies", {}).items():
            dev_dependencies.setdefault(dev_group, []).extend(items)
        deps_setter = [
            (
                metadata.get("optional-dependencies", {}),
                lambda x: (
                    metadata.setdefault("optional-dependencies", {}).__setitem__(group, x)
                    if x
                    else metadata.setdefault("optional-dependencies", {}).pop(group, None)
                ),
            ),
            (dev_dependencies, update_dev_dependencies),
        ]
        normalized_group = normalize_name(group)
        for deps, setter in deps_setter:
            normalized_groups = {normalize_name(g) for g in deps}
            if group in deps:
                return make_array(deps[group], True), setter
            if normalized_group in normalized_groups:
                raise PdmUsageError(f"Group {group} already exists in another non-normalized form")
        # If not found, return an empty list and a setter to add the group
        return tomlkit.array().multiline(True), deps_setter[int(dev)][1]

    def add_dependencies(
        self,
        requirements: Iterable[str | Requirement],
        to_group: str = "default",
        dev: bool = False,
        show_message: bool = True,
        write: bool = True,
    ) -> list[Requirement]:
        """Add requirements to the given group, and return the requirements of that group."""
        if isinstance(requirements, Mapping):  # pragma: no cover
            deprecation_warning(
                "Passing a requirements map to add_dependencies is deprecated, please pass an iterable", stacklevel=2
            )
            requirements = requirements.values()
        self.pyproject.open_for_write()
        deps, setter = self.use_pyproject_dependencies(to_group, dev)
        updated_indices: set[int] = set()

        with cd(self.root):
            parsed_deps = [(parse_line(dep) if isinstance(dep, str) else None) for dep in deps]

            for req in requirements:
                if isinstance(req, str):
                    req = parse_line(req)
                matched_index = next(
                    (
                        i
                        for i, r in enumerate(deps)
                        if isinstance(r, str) and req.matches(r) and i not in updated_indices
                    ),
                    None,
                )
                dep = req.as_line()
                if matched_index is None:
                    updated_indices.add(len(deps))
                    deps.append(dep)
                    parsed_deps.append(req)
                else:
                    deps[matched_index] = dep
                    parsed_deps[matched_index] = req
                    updated_indices.add(matched_index)
        setter(deps)
        if write:
            self.pyproject.write(show_message)
        for r in parsed_deps:
            if r is not None:
                r.groups = [to_group]
        return [r for r in parsed_deps if r is not None]

    def init_global_project(self) -> None:
        if not self.is_global or not self.pyproject.empty():
            return
        self.root.mkdir(parents=True, exist_ok=True)
        self.pyproject.set_data({"project": {"dependencies": ["pip", "setuptools", "wheel"]}})
        self.pyproject.write()

    @property
    def backend(self) -> BuildBackend:
        return get_backend_by_spec(self.pyproject.build_system)(self.root)

    def cache(self, name: str) -> Path:
        path = self.cache_dir / name
        try:
            path.mkdir(parents=True, exist_ok=True)
        except OSError:
            # The path could be not accessible
            pass
        return path

    def make_wheel_cache(self) -> WheelCache:
        from pdm.models.caches import get_wheel_cache

        return get_wheel_cache(self.cache("wheels"))

    @property
    def package_cache(self) -> PackageCache:
        return PackageCache(self.cache("packages"))

    def make_candidate_info_cache(self) -> CandidateInfoCache:
        from pdm.models.caches import CandidateInfoCache, EmptyCandidateInfoCache

        python_hash = hashlib.sha1(str(self.environment.python_requires).encode()).hexdigest()
        file_name = f"package_meta_{python_hash}.json"
        return (
            CandidateInfoCache(self.cache("metadata") / file_name)
            if self.core.state.enable_cache
            else EmptyCandidateInfoCache(self.cache("metadata") / file_name)
        )

    def make_hash_cache(self) -> HashCache:
        from pdm.models.caches import EmptyHashCache, HashCache

        return HashCache(self.cache("hashes")) if self.core.state.enable_cache else EmptyHashCache(self.cache("hashes"))

    def iter_interpreters(
        self,
        python_spec: str | None = None,
        search_venv: bool | None = None,
        filter_func: Callable[[PythonInfo], bool] | None = None,
        respect_version_file: bool = True,
    ) -> Iterable[PythonInfo]:
        """Iterate over all interpreters that matches the given specifier.
        And optionally install the interpreter if not found.
        """
        from packaging.version import InvalidVersion

        from pdm.cli.commands.python import InstallCommand

        def read_version_from_version_file(python_version_file: Path) -> str | None:
            content = python_version_file.read_text().strip()
            content_lines = [cl for cl in content.splitlines() if not cl.lstrip().startswith("#")]

            return content_lines[0] if len(content_lines) == 1 else None

        version_file = self.root.joinpath(".python-version")
        found = False
        if respect_version_file and not python_spec and (os.getenv("PDM_PYTHON_VERSION") or version_file.exists()):
            requested = os.getenv("PDM_PYTHON_VERSION") or read_version_from_version_file(version_file)
            if requested is not None and requested not in self.python_requires:
                self.core.ui.warn(".python-version is found but the version is not in requires-python, ignored.")
            elif requested is not None:
                python_spec = requested
        for interpreter in self.find_interpreters(python_spec, search_venv):
            if filter_func is None or filter_func(interpreter):
                found = True
                yield interpreter
        if found or self.is_global:
            return

        if not python_spec:  # handle both empty string and None
            # Get the best match meeting the requires-python
            best_match = self.get_best_matching_cpython_version()
            if best_match is None:
                return
            python_spec = str(best_match)
        else:
            try:
                if python_spec not in self.python_requires:
                    return
            except InvalidVersion:
                return
        try:
            # otherwise if no interpreter is found, try to install it
            installed = InstallCommand.install_python(self, python_spec)
        except Exception as e:
            self.core.ui.error(f"Failed to install Python {python_spec}: {e}")
            return
        else:
            if filter_func is None or filter_func(installed):
                yield installed

    def find_interpreters(
        self, python_spec: str | None = None, search_venv: bool | None = None
    ) -> Iterable[PythonInfo]:
        """Return an iterable of interpreter paths that matches the given specifier,
        which can be:
            1. a version specifier like 3.7
            2. an absolute path
            3. a short name like python3
            4. None that returns all possible interpreters
        """
        config = self.config
        python: str | Path | None = None
        finder_arg: str | None = None

        if not python_spec:
            if config.get("python.use_pyenv", True) and os.path.exists(PYENV_ROOT):
                pyenv_shim = os.path.join(PYENV_ROOT, "shims", "python3")
                if os.name == "nt":
                    pyenv_shim += ".bat"
                if os.path.exists(pyenv_shim):
                    yield PythonInfo.from_path(pyenv_shim)
                elif os.path.exists(pyenv_shim.replace("python3", "python")):
                    yield PythonInfo.from_path(pyenv_shim.replace("python3", "python"))
            python = shutil.which("python") or shutil.which("python3")
            if python:
                yield PythonInfo.from_path(python)
        else:
            if not all(c.isdigit() for c in python_spec.split(".")):
                path = Path(python_spec)
                if path.exists():
                    python = find_python_in_path(python_spec)
                    if python:
                        yield PythonInfo.from_path(python)
                        return
                if len(path.parts) == 1:  # only check for spec with only one part
                    python = shutil.which(python_spec)
                    if python:
                        yield PythonInfo.from_path(python)
                        return
            finder_arg = python_spec
        if search_venv is None:
            search_venv = cast(bool, config["python.use_venv"])
        finder = self._get_python_finder(search_venv)
        for entry in finder.find_all(finder_arg, allow_prereleases=True):
            yield PythonInfo(entry)
        if not python_spec:
            # Lastly, return the host Python as well
            this_python = getattr(sys, "_base_executable", sys.executable)
            yield PythonInfo.from_path(this_python)

    def _get_python_finder(self, search_venv: bool = True) -> Finder:
        from findpython import Finder

        from pdm.cli.commands.venv.utils import VenvProvider

        providers: list[str] = self.config["python.providers"]
        venv_pos = -1
        if not providers:
            venv_pos = 0
        elif "venv" in providers:
            venv_pos = providers.index("venv")
            providers.remove("venv")
        old_rye_root = os.getenv("RYE_PY_ROOT")
        os.environ["RYE_PY_ROOT"] = os.path.expanduser(self.config["python.install_root"])
        try:
            finder = Finder(resolve_symlinks=True, selected_providers=providers or None)
        finally:
            if old_rye_root:  # pragma: no cover
                os.environ["RYE_PY_ROOT"] = old_rye_root
            else:
                del os.environ["RYE_PY_ROOT"]
        if search_venv and venv_pos >= 0:
            finder.add_provider(VenvProvider(self), venv_pos)
        return finder

    @property
    def is_distribution(self) -> bool:
        if not self.name:
            return False
        settings = self.pyproject.settings
        if "package-type" in settings:
            return settings["package-type"] == "library"
        elif "distribution" in settings:
            return cast(bool, settings["distribution"])
        else:
            return True

    def get_setting(self, key: str) -> Any:
        """
        Get a setting from its dotted key (without the `tool.pdm` prefix).

        Returns `None` if the key does not exists.
        """
        try:
            return reduce(operator.getitem, key.split("."), self.pyproject.settings)
        except KeyError:
            return None

    def env_or_setting(self, var: str, key: str) -> Any:
        """
        Get a value from environment variable and fallback on a given setting.

        Returns `None` if both the environment variable and the key does not exists.
        """
        return os.getenv(var.upper()) or self.get_setting(key)

    def get_best_matching_cpython_version(
        self, use_minimum: bool | None = False, freethreaded: bool = False
    ) -> PythonVersion | None:
        """
        Returns the best matching CPython version that fits requires-python, this platform and arch.
        If no best match could be found, return None.

        Default for best match strategy is "highest" possible interpreter version. If "minimum" shall be used,
        set `use_minimum` to True.
        """

        def get_version(version: PythonVersion) -> str:
            return f"{version.major}.{version.minor}.{version.micro}"

        all_matches = get_all_installable_python_versions(build_dir=False)
        filtered_matches = [
            v
            for v in all_matches
            if v.freethreaded == freethreaded
            and get_version(v) in self.python_requires
            and v.implementation.lower() == "cpython"
        ]
        if filtered_matches:
            if use_minimum:
                return min(filtered_matches, key=lambda v: (v.major, v.minor, v.micro))
            return max(filtered_matches, key=lambda v: (v.major, v.minor, v.micro))

        return None

    @property
    def lock_targets(self) -> list[EnvSpec]:
        return [self.environment.allow_all_spec]

    def get_resolver(self, allow_uv: bool = True) -> type[Resolver]:
        """Get the resolver class to use for the project."""
        from pdm.resolver.resolvelib import RLResolver
        from pdm.resolver.uv import UvResolver

        if allow_uv and self.config.get("use_uv"):
            return UvResolver
        else:
            return RLResolver

    def get_synchronizer(self, quiet: bool = False, allow_uv: bool = True) -> type[BaseSynchronizer]:
        """Get the synchronizer class to use for the project."""
        from pdm.installers import BaseSynchronizer, Synchronizer, UvSynchronizer
        from pdm.installers.uv import QuietUvSynchronizer

        if allow_uv and self.config.get("use_uv"):
            return QuietUvSynchronizer if quiet else UvSynchronizer
        if quiet:
            return BaseSynchronizer
        return getattr(self.core, "synchronizer_class", Synchronizer)



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, Callable

import unearth
from dep_logic.specifiers import InvalidSpecifier, parse_version_specifier
from packaging.version import Version
from unearth.evaluator import Evaluator, FormatControl, LinkMismatchError, Package

from pdm.models.markers import EnvSpec
from pdm.utils import parse_version

logger = logging.getLogger("unearth")

if TYPE_CHECKING:
    from pdm.models.session import PDMPyPIClient


class ReverseVersion(Version):
    """A subclass of version that reverse the order of comparison."""

    def __lt__(self, other: Any) -> bool:
        return super().__gt__(other)

    def __le__(self, other: Any) -> bool:
        return super().__ge__(other)

    def __gt__(self, other: Any) -> bool:
        return super().__lt__(other)

    def __ge__(self, other: Any) -> bool:
        return super().__le__(other)


class PDMEvaluator(Evaluator):
    def __init__(self, *args: Any, env_spec: EnvSpec, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.env_spec = env_spec

    def check_requires_python(self, link: unearth.Link) -> None:
        if link.requires_python:
            try:
                requires_python = parse_version_specifier(link.requires_python)
            except InvalidSpecifier as e:
                logger.debug(
                    "Invalid requires-python specifier for link(%s) %s: %s", link.redacted, link.requires_python, e
                )
                return
            if (requires_python & self.env_spec.requires_python).is_empty():
                raise LinkMismatchError(
                    f"The package requires-python {link.requires_python} is not compatible with the target {self.env_spec.requires_python}."
                )

    def check_wheel_tags(self, filename: str) -> None:
        if self.env_spec.wheel_compatibility(filename) is None:
            raise LinkMismatchError(
                f"The wheel file {filename} is not compatible with the target environment {self.env_spec}."
            )


class PDMPackageFinder(unearth.PackageFinder):
    def __init__(
        self,
        session: PDMPyPIClient | None = None,
        *,
        env_spec: EnvSpec,
        minimal_version: bool = False,
        **kwargs: Any,
    ) -> None:
        super().__init__(session, **kwargs)
        self.minimal_version = minimal_version
        self.env_spec = env_spec

    def build_evaluator(self, package_name: str, allow_yanked: bool = False) -> Evaluator:
        format_control = FormatControl(no_binary=self.no_binary, only_binary=self.only_binary)
        return PDMEvaluator(
            package_name=package_name,
            target_python=self.target_python,
            allow_yanked=allow_yanked,
            format_control=format_control,
            exclude_newer_than=self.exclude_newer_than,
            env_spec=self.env_spec,
        )

    def _sort_key(self, package: Package) -> tuple:
        from packaging.utils import BuildTag, canonicalize_name

        if self.minimal_version:
            version_cls: Callable[[str], Version] = ReverseVersion
        else:
            version_cls = parse_version

        link = package.link
        compatibility = (0, 0, 0, 0)  # default value for sdists
        build_tag: BuildTag = ()
        prefer_binary = False
        if link.is_wheel:
            compat = self.env_spec.wheel_compatibility(link.filename)
            if compat is None:
                compatibility = (-1, -1, -1, -1)
            else:
                compatibility = compat
            if canonicalize_name(package.name) in self.prefer_binary or ":all:" in self.prefer_binary:
                prefer_binary = True

        return (
            -int(link.is_yanked),
            int(prefer_binary),
            version_cls(package.version) if package.version is not None else version_cls("0"),
            compatibility,
            build_tag,
        )



from pdm.environments.base import BareEnvironment, BaseEnvironment
from pdm.environments.local import PythonLocalEnvironment
from pdm.environments.python import PythonEnvironment

__all__ = [
    "BareEnvironment",
    "BaseEnvironment",
    "PythonEnvironment",
    "PythonLocalEnvironment",
]



from __future__ import annotations

import functools
import traceback
from concurrent.futures import Future, ThreadPoolExecutor
from types import SimpleNamespace
from typing import TYPE_CHECKING

from pdm import termui
from pdm.exceptions import InstallationError
from pdm.installers.base import BaseSynchronizer
from pdm.models.candidates import Candidate
from pdm.models.reporter import CandidateReporter, InstallationStatus, RichProgressReporter
from pdm.models.requirements import strip_extras

if TYPE_CHECKING:
    from rich.progress import Progress

    from pdm.compat import Distribution


class Synchronizer(BaseSynchronizer):
    def install_candidate(self, key: str, progress: Progress) -> Candidate:
        """Install candidate"""
        can = self.candidates[key]
        job = progress.add_task(f"Installing {can.format()}...", text="", total=None)
        can.prepare(self.environment, RichProgressReporter(progress, job))
        try:
            self.manager.install(can)
        except Exception:
            progress.print(f"  [error]{termui.Emoji.FAIL}[/] Install {can.format()} failed")
            raise
        else:
            progress.print(f"  [success]{termui.Emoji.SUCC}[/] Install {can.format()} successful")
        finally:
            progress.remove_task(job)
            can.prepare(self.environment, CandidateReporter())
        return can

    def update_candidate(self, key: str, progress: Progress) -> tuple[Distribution, Candidate]:
        """Update candidate"""
        can = self.candidates[key]
        dist = self.working_set[strip_extras(key)[0]]
        dist_version = dist.version
        job = progress.add_task(
            f"Updating [req]{key}[/] [warning]{dist_version}[/] -> [warning]{can.version}[/]...", text="", total=None
        )
        can.prepare(self.environment, RichProgressReporter(progress, job))
        try:
            self.manager.overwrite(dist, can)
        except Exception:
            progress.print(
                f"  [error]{termui.Emoji.FAIL}[/] Update [req]{key}[/] "
                f"[warning]{dist_version}[/] "
                f"-> [warning]{can.version}[/] failed",
            )
            raise
        else:
            progress.print(
                f"  [success]{termui.Emoji.SUCC}[/] Update [req]{key}[/] "
                f"[warning]{dist_version}[/] "
                f"-> [warning]{can.version}[/] successful",
            )
        finally:
            progress.remove_task(job)
            can.prepare(self.environment, CandidateReporter())

        return dist, can

    def remove_distribution(self, key: str, progress: Progress) -> Distribution:
        """Remove distributions with given names."""
        dist = self.working_set[key]
        dist_version = dist.version

        job = progress.add_task(f"Removing [req]{key}[/] [warning]{dist_version}[/]...", text="", total=None)
        try:
            self.manager.uninstall(dist)
        except Exception:
            progress.print(
                f"  [error]{termui.Emoji.FAIL}[/] Remove [req]{key}[/] [warning]{dist_version}[/] failed",
            )
            raise
        else:
            progress.print(
                f"  [success]{termui.Emoji.SUCC}[/] Remove [req]{key}[/] [warning]{dist_version}[/] successful"
            )
        finally:
            progress.remove_task(job)
        return dist

    def _show_headline(self, packages: dict[str, list[str]]) -> None:
        add, update, remove = packages["add"], packages["update"], packages["remove"]
        if not any((add, update, remove)):
            self.ui.echo("All packages are synced to date, nothing to do.")
            return
        results = ["[bold]Synchronizing working set with resolved packages[/]:"]
        results.extend(
            [
                f"[success]{len(add)}[/] to add,",
                f"[warning]{len(update)}[/] to update,",
                f"[error]{len(remove)}[/] to remove",
            ]
        )
        self.ui.echo(" ".join(results) + "\n")

    def _show_summary(self, packages: dict[str, list[str]]) -> None:
        to_add = [self.candidates[key] for key in packages["add"]]
        to_update = [(self.working_set[key], self.candidates[key]) for key in packages["update"]]
        to_remove = [self.working_set[key] for key in packages["remove"]]
        lines = []
        if to_add:
            lines.append("[bold]Packages to add[/]:")
            for can in to_add:
                lines.append(f"  - {can.format()}")
        if to_update:
            lines.append("[bold]Packages to update[/]:")
            for prev, cur in to_update:
                lines.append(f"  - [req]{cur.name}[/] [warning]{prev.version}[/] -> [warning]{cur.version}[/]")
        if to_remove:
            lines.append("[bold]Packages to remove[/]:")
            for dist in to_remove:
                lines.append(f"  - [req]{dist.metadata['Name']}[/] [warning]{dist.version}[/]")
        if lines:
            self.ui.echo("\n".join(lines))
        else:
            self.ui.echo("All packages are synced to date, nothing to do.")

    def _fix_pth_files(self) -> None:
        """Remove the .pdmtmp suffix from the installed packages"""
        from pathlib import Path

        lib_paths = self.environment.get_paths()
        for scheme in ["purelib", "platlib"]:
            if not Path(lib_paths[scheme]).exists():
                continue
            for path in list(Path(lib_paths[scheme]).iterdir()):
                if path.suffix == ".pdmtmp":
                    target_path = path.with_suffix("")
                    if target_path.exists():
                        target_path.unlink()
                    path.rename(target_path)

    def synchronize(self) -> None:
        to_add, to_update, to_remove = self.compare_with_working_set()
        to_do = {"remove": to_remove, "update": to_update, "add": to_add}

        if self.dry_run:
            self._show_summary(to_do)
            return

        self._show_headline(to_do)
        handlers = {
            "add": self.install_candidate,
            "update": self.update_candidate,
            "remove": self.remove_distribution,
        }
        sequential_jobs = []
        parallel_jobs = []

        for kind in to_do:
            for key in to_do[kind]:
                if key in self.SEQUENTIAL_PACKAGES or not self.parallel:
                    sequential_jobs.append((kind, key))
                elif key in self.candidates and self.candidates[key].req.editable:
                    # Editable packages are installed sequentially.
                    sequential_jobs.append((kind, key))
                else:
                    parallel_jobs.append((kind, key))

        state = SimpleNamespace(errors=[], parallel_failed=[], sequential_failed=[], jobs=[], mark_failed=False)

        def update_progress(future: Future, kind: str, key: str) -> None:
            error = future.exception()
            status.update_spinner(advance=1)  # type: ignore[has-type]
            if error:
                exc_info = (type(error), error, error.__traceback__)
                termui.logger.exception("Error occurs %sing %s: ", kind.rstrip("e"), key, exc_info=exc_info)
                state.parallel_failed.append((kind, key))
                state.errors.extend([f"{kind} [success]{key}[/] failed:\n", *traceback.format_exception(*exc_info)])
                if self.fail_fast:
                    for future in state.jobs:
                        future.cancel()
                    state.mark_failed = True

        # get rich progress and live handler to deal with multiple spinners
        with InstallationStatus(self.ui, "Synchronizing") as status:
            for i in range(self.retry_times + 1):
                status.update_spinner(completed=0, total=len(sequential_jobs) + len(parallel_jobs))
                for kind, key in sequential_jobs:
                    try:
                        handlers[kind](key, status.progress)
                    except Exception:
                        termui.logger.exception("Error occurs: ")
                        state.sequential_failed.append((kind, key))
                        state.errors.extend([f"{kind} [success]{key}[/] failed:\n", traceback.format_exc()])
                        if self.fail_fast:
                            state.mark_failed = True
                            break
                    finally:
                        status.update_spinner(advance=1)
                if state.mark_failed:
                    break
                state.jobs.clear()
                if parallel_jobs:
                    with ThreadPoolExecutor() as executor:
                        for kind, key in parallel_jobs:
                            future = executor.submit(handlers[kind], key, status.progress)
                            future.add_done_callback(functools.partial(update_progress, kind=kind, key=key))
                            state.jobs.append(future)
                if (
                    state.mark_failed
                    or i == self.retry_times
                    or (not state.sequential_failed and not state.parallel_failed)
                ):
                    break
                sequential_jobs, state.sequential_failed = state.sequential_failed, []
                parallel_jobs, state.parallel_failed = state.parallel_failed, []
                state.errors.clear()
                status.update_spinner(description=f"Retry failed jobs({i + 2}/{self.retry_times + 1})")

            try:
                if state.errors:
                    if self.ui.verbosity < termui.Verbosity.DETAIL:
                        status.console.print("\n[error]ERRORS[/]:")
                        status.console.print("".join(state.errors), end="")
                    status.update_spinner(description=f"[error]{termui.Emoji.FAIL}[/] Some package operations failed.")
                    raise InstallationError("Some package operations failed.")

                if self.install_self:
                    self_key = self.self_key
                    assert self_key
                    self.candidates[self_key] = self.self_candidate
                    word = "a" if self.no_editable else "an editable"
                    status.update_spinner(description=f"Installing the project as {word} package...")
                    if self_key in self.working_set:
                        self.update_candidate(self_key, status.progress)
                    else:
                        self.install_candidate(self_key, status.progress)

                status.update_spinner(description=f"{termui.Emoji.POPPER} All complete!")
            finally:
                # Now we remove the .pdmtmp suffix from the installed packages
                self._fix_pth_files()



from __future__ import annotations

import argparse
import textwrap

from pdm import termui
from pdm._types import SearchResults
from pdm.cli.commands.base import BaseCommand
from pdm.cli.options import verbose_option
from pdm.models.working_set import WorkingSet
from pdm.project import Project
from pdm.utils import normalize_name


def print_results(
    ui: termui.UI,
    hits: SearchResults,
    working_set: WorkingSet,
    terminal_width: int | None = None,
) -> None:
    if not hits:
        return
    name_column_width = max(len(hit.name) + len(hit.version or "") for hit in hits) + 4

    for hit in hits:
        name = hit.name
        summary = hit.summary or ""
        if terminal_width is not None:
            target_width = terminal_width - name_column_width - 5
            if target_width > 10:
                # wrap and indent summary to fit terminal
                summary = ("\n" + " " * (name_column_width + 2)).join(textwrap.wrap(summary, target_width))
        current_width = len(name) + 1
        spaces = " " * (name_column_width - current_width)
        line = f"[req]{name}[/]{spaces} - {summary}"
        try:
            ui.echo(line)
            if normalize_name(name) in working_set:
                dist = working_set[normalize_name(name)]
                ui.echo(f"  INSTALLED: {dist.version}")
        except UnicodeEncodeError:
            pass


class Command(BaseCommand):
    """[DEPRECATED] Search for PyPI packages"""

    arguments = (verbose_option,)

    def add_arguments(self, parser: argparse.ArgumentParser) -> None:
        parser.add_argument("query", help="Query string to search")

    def handle(self, project: Project, options: argparse.Namespace) -> None:
        project.core.ui.warn(
            "Since pypi.org no longer supports search API, this command is deprecated and will be removed in future versions. "
            "Please visit `https://pypi.org` in the browser to search for packages.",
        )
        return
